from promnesia.common import Source
from promnesia.sources import auto

'''
List of sources to use.

You can specify your own, add more sources, etc.
See https://github.com/karlicoss/promnesia#setup for more information
'''
SOURCES = [
    Source(
        auto.index,
        # just some arbitrary directory with plaintext files
        '/usr/include/c++/',
        '/usr/local/include/c++/', # on apple they are here apparently..
    )
]



#!/usr/bin/env python3
import argparse
import os
import sys
import time
from pathlib import Path
import platform
from subprocess import check_call, run
from typing import List

SYSTEM = platform.system()
UNSUPPORTED_SYSTEM = RuntimeError(f'Platform {SYSTEM} is not supported yet!')
NO_SYSTEMD = RuntimeError('systemd not detected, find your own way to start promnesia automatically')

from ..common import root
from ..server import setup_parser as server_setup_parser

SYSTEMD_TEMPLATE = '''
[Unit]
Description=Promnesia browser extension backend

[Install]
WantedBy=default.target

[Service]
ExecStart={launcher} {extra_args}
Type=simple
Restart=always
'''

LAUNCHD_TEMPLATE = '''
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
        <dict>
                <key>Label</key>
                <string>{service_name}</string>

                <key>ProgramArguments</key>
                <array>
{arguments}
                </array>

                <key>RunAtLoad</key>
                <true/>
                <key>KeepAlive</key>
                <true/>
        </dict>
</plist>
'''


def systemd(*args, method=check_call):
    method([
        'systemctl', '--no-pager', '--user', *args,
    ])


def install_systemd(name: str, out: Path, launcher: str, largs: List[str]) -> None:
    unit_name = name

    import shlex
    extra_args = ' '.join(shlex.quote(str(a)) for a in largs)

    out.write_text(SYSTEMD_TEMPLATE.format(
        launcher=launcher,
        extra_args=extra_args,
    ))

    try:
        systemd('stop' , unit_name, method=run) # ignore errors here if it wasn't running in the first place
        systemd('daemon-reload')
        systemd('enable', unit_name)
        systemd('start' , unit_name)
        systemd('status', unit_name)
    except Exception as e:
        print(f"Something has gone wrong... you might want to use 'journalctl --user -u {unit_name}' to investigate", file=sys.stderr)
        raise e


def install_launchd(name: str, out: Path, launcher: str, largs: List[str]) -> None:
    service_name = name
    arguments = '\n'.join(f'<string>{a}</string>' for a in [launcher, *largs])
    out.write_text(LAUNCHD_TEMPLATE.format(
        service_name=service_name,
        arguments=arguments,
    ))
    cmd = ['launchctl', 'load', '-w', str(out)]
    print('Running: ' + ' '.join(cmd), file=sys.stderr)
    check_call(cmd)

    time.sleep(1) # to give it some time? not sure if necessary
    check_call(f'launchctl list | grep {name}', shell=True)


def install(args) -> None:
    name = args.name
    # todo use appdirs for config dir detection
    if SYSTEM == 'Linux':
        # Check for existence of systemd
        # https://www.freedesktop.org/software/systemd/man/sd_booted.html
        if not Path('/run/systemd/system/').exists():
            raise NO_SYSTEMD
        suf = '.service'
        if Path(name).suffix != suf:
            name = name + suf
        out = Path(f'~/.config/systemd/user/{name}')
    elif SYSTEM == 'Darwin': # osx
        out = Path(f'~/Library/LaunchAgents/{name}.plist')
    else:
        raise UNSUPPORTED_SYSTEM
    out = out.expanduser()
    print(f"Writing launch script to {out}", file=sys.stderr)

    # ugh. we want to know whether we're invoked 'properly' as an executable or ad-hoc via scripts/promnesia
    if os.environ.get('DIRTY_RUN') is not None:
        launcher = str(root() / 'scripts/promnesia')
    else:
        # must be installed, so available in PATH
        import distutils.spawn
        exe = distutils.spawn.find_executable('promnesia'); assert exe is not None
        launcher = exe # older systemd wants absolute paths..

    db = args.db
    largs = [
        'serve',
        *([] if db is None else ['--db', str(db)]),
        '--timezone', args.timezone,
        '--port', args.port,
    ]

    out.parent.mkdir(parents=True, exist_ok=True) # sometimes systemd dir doesn't exist
    if SYSTEM == 'Linux':
        install_systemd(name=name, out=out, launcher=launcher, largs=largs)
    elif SYSTEM == 'Darwin':
        install_launchd(name=name, out=out, launcher=launcher, largs=largs)
    else:
        raise UNSUPPORTED_SYSTEM


def setup_parser(p: argparse.ArgumentParser) -> None:
    if SYSTEM == 'Linux':
        dflt = 'promnesia.service'
    elif SYSTEM == 'Darwin':
        dflt = 'com.github.karlicoss.promnesia'
    else:
        # defensive here because setup_parser is called regardless whether the functionality is used
        dflt = NotImplemented

    p.add_argument('--name', type=str, default=dflt, help='Systemd/launchd service name')
    p.add_argument('--unit-name', type=str, dest='name', help='DEPRECATED, same as --name')
    server_setup_parser(p)



#!/usr/bin/env python3
from contextlib import contextmanager
from datetime import datetime, date
from os.path import join, getsize
from pathlib import Path
from tempfile import TemporaryDirectory
import pytz
from shutil import copytree
import os
from os import mkdir
from os.path import lexists
from typing import Union, List

import pytest # type: ignore
from pytest import mark # type: ignore


from common import tdata, reset_hpi_modules
from config_tests import with_config

from promnesia.common import Visit, Indexer, Loc, Res, DbVisit, _is_windows

# TODO need to expire dbcache in tests..

skip = mark.skip


def W(*args, **kwargs):
    if 'src' not in kwargs:
        kwargs['src'] = 'whatever'
    return Indexer(*args, **kwargs)


def as_visits(*args, **kwargs) -> List[Res[DbVisit]]:
    from promnesia.extract import extract_visits
    kwargs['src'] = 'whatever'
    return list(extract_visits(*args, **kwargs))


def as_ok_visits(*args, **kwargs) -> List[DbVisit]:
    r: List[DbVisit] = []
    for v in as_visits(*args, **kwargs):
        if isinstance(v, Exception):
            raise v
        r.append(v)
    return r


from contextlib import contextmanager
@contextmanager
def extra_path(p: Path):
    import sys
    try:
        sys.path.append(str(p))
        yield
    finally:
        sys.path.pop()

def import_file(p: Union[str, Path], name=None):
    p = Path(p)
    if name is None:
        name = p.stem
    import importlib.util
    spec = importlib.util.spec_from_file_location(name, p); assert spec is not None
    foo = importlib.util.module_from_spec(spec)
    with extra_path(p.parent):
        spec.loader.exec_module(foo) # type: ignore
    return foo


# TODO eh. need to separate stuff for history backups out...
backup_db = import_file('scripts/browser_history.py')

def assert_got_tzinfo(visits):
    for v in visits:
        assert v.dt.tzinfo is not None


# TODO I guess global get_config methods is ok? command line can populate it, also easy to hack in code?
# TODO cache should be in the configuration I suppose?

@pytest.fixture
def adhoc_config(tmp_path: Path):
    cdir = tmp_path / 'cache'
    cdir.mkdir()

    from promnesia import config

    try:
        config.instance = config.Config(
            OUTPUT_DIR=tmp_path,
            CACHE_DIR=cdir,
        )
        yield
    finally:
        config.reset()


def test_with_error() -> None:
    class ExtractionError(Exception):
        pass
    def err_ex():
        for i in range(3):
            if i == 1:
                yield ExtractionError()
            else:
                yield Visit(
                    url=f'http://test{i}',
                    dt=datetime.utcfromtimestamp(0),
                    locator=Loc.make('whatever'),
                )
    [v1, e, v2] = as_visits(lambda: err_ex())
    assert isinstance(v1, DbVisit)
    assert isinstance(e, Exception)
    assert isinstance(v2, DbVisit)


# todo testing this logic probably belongs to hpi or google_takeout_export, but whatever
def test_takeout_directory(adhoc_config, tmp_path: Path) -> None:
    reset_hpi_modules()
    from my.cfg import config
    class user_config:
        takeout_path = tdata('takeout')
    config.google = user_config # type: ignore

    # TODO ugh, the disabled_cachew thing isn't very nice
    from my.core.cachew import disabled_cachew
    with disabled_cachew():
        import promnesia.sources.takeout as tex
        visits = as_ok_visits(W(tex.index))
    assert len(visits) == 3

    assert_got_tzinfo(visits)


def test_takeout_zip(adhoc_config) -> None:
    reset_hpi_modules()
    from my.cfg import config
    class user_config:
        takeout_path = tdata('takeout-20150518T000000Z.zip')
    config.google = user_config # type: ignore

    from my.core.cachew import disabled_cachew
    with disabled_cachew():
        import promnesia.sources.takeout as tex
        visits = as_ok_visits(tex.index)
    assert len(visits) == 3
    [vis] = [v for v in visits if v.norm_url == 'takeout.google.com/settings/takeout']

    edt = datetime(
        year=2018,
        month=9,
        day=18,
        hour=5,
        minute=48,
        second=23,
        tzinfo=pytz.utc,
    )
    assert vis.dt == edt

    assert_got_tzinfo(visits)


@skip("TODO not sure how to trigger firefox on CI...")
def test_backup_firefox(tmp_path):
    tdir = Path(tmp_path)
    path = backup_db.backup_history('firefox', to=tdir, profile='*release*')
    # shouldn't fail at least

        # [hist] = list(chrome_gen.iter_chrome_histories(path, 'sqlite'))
        # assert len(hist) > 10 # kinda random sanity check

        # render([hist], join(tdir, 'res.json'))

        # assert_got_tzinfo(hist)


def test_plaintext_path_extractor() -> None:
    import promnesia.sources.shellcmd as custom_gen
    from promnesia.sources.plaintext import extract_from_path

    visits = as_ok_visits(W(custom_gen.index,
        extract_from_path(tdata('custom')),
    ))
    assert {
        v.orig_url for v in visits
    } == {
        'http://google.com',
        'http://google.com/',
        'http://some-weird-domain.xyz/whatever',
        'https://google.com',
        'http://what.about.this.link',
    }

    [wa] = [v for v in visits if v.orig_url == 'http://what.about.this.link']
    f2 = Path(tdata('custom')) / 'file2.txt'
    assert wa.locator.href == f'editor://{f2}:3' # occurs line 3

# TODO perhaps it belongs to canonify?
def test_normalise() -> None:
    import promnesia.sources.shellcmd as custom_gen
    from promnesia.sources.plaintext import extract_from_path

    visits = as_ok_visits(W(custom_gen.index,
        extract_from_path(tdata('normalise')),
    ))
    assert len(visits) == 7
    assert {
        v.norm_url for v in visits
    } == {
        'hi.com',
        'reddit.com/post',
        'argos.co.uk/webapp/wcs/stores/servlet/OrderItemDisplay',
        'youtube.com/watch?v=XXlZfc1TrD0',
        'youtube.com/watch?v=XXlZfc1Tr11',
    }


def test_normalise_weird() -> None:
    import promnesia.sources.shellcmd as custom_gen
    from promnesia.sources.plaintext import extract_from_path

    visits = as_ok_visits(W(
        custom_gen.index,
        extract_from_path(tdata('weird.txt')),
    ))
    [v1, v2] = visits

    # TODO assert there are no spaces in the database?
    assert "urbandictionary.com/define.php?term=Belgian%20Whistle" == v1.norm_url

    assert "en.wikipedia.org/wiki/Dinic%27s_algorithm"             == v2.norm_url
    assert v2.locator.title.endswith('weird.txt:2')
    assert v2.context == 'right, so https://en.wikipedia.org/wiki/Dinic%27s_algorithm can be used for max flow'


def test_filter() -> None:
    import promnesia.sources.shellcmd as custom_gen
    from promnesia.sources.plaintext import extract_from_path

    # ugh... such a mess
    @contextmanager
    def reset_filters():
        try:
            E.filters.cache_clear()
            yield
        finally:
            E.filters.cache_clear()

    import promnesia.extract as E
    with reset_filters(), with_config('''
FILTERS = [
    "some-weird-domain.xyz"
]
'''):
        visits = as_visits(W(
            custom_gen.index,
            extract_from_path(tdata('custom')),
        ))
        assert len(visits) == 4


@pytest.mark.skipif(_is_windows, reason="no grep on windows")
def test_custom() -> None:
    import promnesia.sources.shellcmd as custom_gen

    visits = as_visits(W(
        custom_gen.index,
        # meh. maybe should deprecate plain string here...
        """grep -Eo -r --no-filename (http|https)://\S+ """ + tdata('custom'),
    ))
    # TODO I guess filtering of equivalent urls should rather be tested on something having context (e.g. org mode)
    assert len(visits) == 5


def test_hook() -> None:
    import promnesia.sources.shellcmd as custom_gen
    from promnesia.__main__ import iter_all_visits
    with with_config('''
from promnesia.common import Source
from promnesia.sources import demo

SOURCES = [
    Source(demo.index, count=7, name='somename'),
]

from typing import Iterable
from promnesia.common import DbVisit, Loc, Res

def HOOK(visit: Res[DbVisit]) -> Iterable[Res[DbVisit]]:
    # NOTE: might be a good idea to check that the visit is an exception first and yield it intact?
    nurl = visit.norm_url
    if 'page1' in nurl:
        yield visit._replace(norm_url='patched.com')
    elif 'page2' in nurl:
        None.boom # deliberately crash
    elif 'page3' in nurl:
        # just don't yield anything! it will be omitted
        pass
    elif 'page4' in nurl:
         # can emit multiple!
        yield visit
        yield visit
    elif 'page6' in nurl:
        # patch locator
        yield visit._replace(locator=Loc.make(title='some custom timte', href='/can/replace/original/path'))
    else:
        yield visit
'''):
        # TODO hmm might be nice to allow in-pace modifications...
        [p0, p1, e2, p41, p42, p5, p6] = list(iter_all_visits())
        assert isinstance(p0, DbVisit)
        assert p0.norm_url == 'demo.com/page0.html'
        assert isinstance(p1, DbVisit)
        assert p1.norm_url == 'patched.com'
        assert isinstance(e2, Exception)
        assert p41 == p42
        assert isinstance(p6, DbVisit)
        assert p6.locator is not None


TESTDATA_CHROME_HISTORY = "/L/data/promnesia/testdata/chrome-history"

def get_chrome_history_backup(td: str):
    copytree(TESTDATA_CHROME_HISTORY, join(td, 'backup'))

@skip("TODO move this to populate script instead")
def test_merge():
    merge = backup_db.merge

    # TODO third is implicit... use merging function
    with TemporaryDirectory() as tdir:
        get_chrome_history_backup(tdir)
        first  = join(tdir, "backup/20180415/History")
        second = join(tdir, "backup/20180417/History")

        mdir = join(tdir, 'merged')
        mkdir(mdir)
        merged_path = join(mdir, 'merged.sql')


        def merged_size() -> int:
            return getsize(merged_path)

        merge(merged_path, first)
        fsize = merged_size()

        merge(merged_path, first)
        fsize_2 = merged_size()

        assert fsize == fsize_2

        merge(merged_path, second)
        ssize = merged_size()

        assert ssize > fsize

        merge(merged_path, second)
        ssize_2 = merged_size()

        assert ssize_2 == ssize


def _test_merge_all_from(tdir):
    merge_all_from = backup_db.merge_all_from # type: ignore
    mdir = join(tdir, 'merged')
    mkdir(mdir)
    mfile = join(mdir, 'merged.sql')

    get_chrome_history_backup(tdir)

    merge_all_from(mfile, join(tdir, 'backup'), None)

    first  = join(tdir, "backup/20180415/History")
    second = join(tdir, "backup/20180417/History")

    # should be removed
    assert not lexists(first)
    assert not lexists(second)

    import promnesia.sources.chrome as chrome_ex # type: ignore

    hist = history(W(chrome_ex.extract, mfile)) # type: ignore
    assert len(hist) > 0

    older = hist['github.com/orgzly/orgzly-android/issues']
    assert any(v.dt.date() < date(year=2018, month=1, day=17) for v in older)
    # in particular, "2018-01-16 19:56:56"

    newer = hist['en.wikipedia.org/wiki/Notice_and_take_down']
    assert any(v.dt.date() >= date(year=2018, month=4, day=16) for v in newer)

    # from implicit db
    newest = hist['feedly.com/i/discover']
    assert any(v.dt.date() >= date(year=2018, month=9, day=27) for v in newest)

@skip("TODO move this to populate script")
def test_merge_all_from(tmp_path):
    tdir = Path(tmp_path)
    _test_merge_all_from(tdir)
    # TODO and also some other unique thing..

if __name__ == '__main__':
    pytest.main([__file__])

# TODO once I integrate test for db population, test chrome/firefox extractors



'''
Uses [[https://github.com/karlicoss/HPI][HPI]] for Pocket highlights & bookmarks
'''
from ..common import Visit, Loc, Results


def index() -> Results:
    from . import hpi
    from my.pocket import articles
    # TODO use docstring from my. module? E.g. describing which pocket format is expected

    for a in articles():
        loc = Loc.make(title='pocket', href=a.pocket_link)
        hls = a.highlights
        if len(hls) == 0:
            yield Visit(
                url=a.url,
                dt=a.added,
                context=None,
                locator=loc,
            )
        for hl in hls:
            yield Visit(
                url=a.url,
                dt=hl.created,
                context=hl.text,
                locator=loc,
            )



from functools import lru_cache
import re
import traceback
from typing import Set, Iterable, Sequence, Union

from .cannon import CanonifyException
from .common import Res, SourceName, DbVisit, Visit, get_logger, Indexer, Filter, Url, Results


logger = get_logger()


DEFAULT_FILTERS = (
    r'^chrome-\w+://',
    r'chrome://newtab',
    r'chrome://apps',
    r'chrome://history',
    r'^about:',
    r'^blob:',
    r'^view-source:',

    r'^content:',
)


@lru_cache(1) #meh, not sure what would happen under tests?
def filters() -> Sequence[Filter]:
    from . import config

    flt = list(DEFAULT_FILTERS)
    if config.has(): # meeeh...
        cfg = config.get()
        flt.extend(cfg.FILTERS)
    return tuple(make_filter(f) for f in flt)


def extract_visits(extractor, *, src: SourceName) -> Iterable[Res[DbVisit]]:
    ex = extractor

    log_info: str
    if isinstance(ex, Indexer):
        log_info = f'{ex.ff.__module__}:{ex.ff.__name__} {ex.args} {ex.kwargs} ...'
        extr = lambda: ex.ff(*ex.args, **ex.kwargs)
    else:
        # TODO if it's a lambda?
        log_info = f'{ex.__module__}:{ex.__name__}'
        extr = ex

    logger.info('extracting via %s ...', log_info)

    try:
        vit: Results = extr()
    except Exception as e:
        # todo critical error?
        logger.exception(e)
        yield e
        return

    handled: Set[Visit] = set()
    try:
        for p in vit:
            if isinstance(p, Exception):
                # todo not sure if need it at all?
                # parts = ['indexer emitted exception\n']
                # eh, exception type is ignored by format_exception completely, apparently??
                # parts.extend(traceback.format_exception(Exception, p, p.__traceback__))
                # logger.error(''.join(parts))
                yield p
                continue

            if p in handled: # no need to emit duplicates
                continue
            handled.add(p)

            yield from as_db_visit(p, src=src)
    except Exception as e:
        # todo critical error?
        logger.exception(e)
        yield e


    logger.info('extracting via %s: got %d visits', log_info, len(handled))


def as_db_visit(v: Visit, *, src: SourceName) -> Iterable[Res[DbVisit]]:
    if filtered(v.url):
        return
    res = DbVisit.make(v, src=src)
    if isinstance(res, CanonifyException):
        # todo not sure if need this log? either way maybe get rid of canonify exception and just yield up
        logger.error('error while canonnifying %s... ignoring', v)
        logger.exception(res)
    yield res


def filtered(url: Url) -> bool:
    return any(f(url) for f in filters())


def make_filter(thing: Union[str, Filter]) -> Filter:
    if isinstance(thing, str):
        rc = re.compile(thing)
        def filter_(u: str) -> bool:
            return rc.search(u) is not None
        return filter_
    else: # must be predicate
        return thing



from pathlib import Path
from typing import Union, Sequence, List, TYPE_CHECKING

PathIsh = Union[Path, str]


Paths = Sequence[PathIsh]

# TLDR: py37 on windows has an annoying bug.. https://github.com/karlicoss/promnesia/issues/91#issuecomment-701051074
def _fix(args: Paths) -> List[str]:
    assert not isinstance(args, str), args # just to prevent shell=True calls...
    return list(map(str, args))


import argparse

def register_argparse_extend_action_in_pre_py38(parser: argparse.ArgumentParser):
    import sys

    if sys.version_info < (3, 8):

        class ExtendAction(argparse.Action):

            def __call__(self, parser, namespace, values, option_string=None):
                items = getattr(namespace, self.dest) or []
                items.extend(values)
                setattr(namespace, self.dest, items)


        parser.register('action', 'extend', ExtendAction)


import subprocess

def run(args: Paths, **kwargs) -> subprocess.CompletedProcess:
    return subprocess.run(_fix(args), **kwargs)

def check_call(args: Paths, **kwargs) -> None:
    subprocess.check_call(_fix(args), **kwargs)

def check_output(args: Paths, **kwargs) -> bytes:
    return subprocess.check_output(_fix(args), **kwargs)

def Popen(args: Paths, **kwargs) -> subprocess.Popen:
    return subprocess.Popen(_fix(args), **kwargs)

PIPE = subprocess.PIPE


# can remove after python3.9
def removeprefix(text: str, prefix: str) -> str:
    if text.startswith(prefix):
        return text[len(prefix):]
    return text



from ..common import Visit, logger, PathIsh, Url, Loc, Results

# TODO make an iterator, insert in db as we go? handle errors gracefully?
def index() -> Results:
    from . import hpi
    from my.google.takeout.paths import get_takeouts
    takeouts = list(get_takeouts())
    # TODO if no takeouts, raise?
    # although could raise a warning on top level, when source emitted no takeouts

    # TODO youtube?
    google_activities = [read_google_activity(t)      for t in takeouts]
    search_activities = [read_search_activity(t)      for t in takeouts]
    browser_histories = [read_browser_history_json(t) for t in takeouts]

    key = lambda v: (v.dt, v.url)
    return chain(
        unique_everseen(chain(*google_activities), key=key),
        unique_everseen(chain(*search_activities), key=key),
        unique_everseen(chain(*browser_histories), key=key),
    )



import pytz
from itertools import chain
from datetime import datetime
from typing import List, Optional, Iterable, TYPE_CHECKING
from pathlib import Path
import json


from .. import config


from more_itertools import unique_everseen
from cachew import cachew


# TODO use CPath? Could encapsulate a path within an archive *or* within a directory
TakeoutPath = Path


def _read_myactivity_html(takeout: TakeoutPath, kind: str) -> Iterable[Visit]:
    from my.core.kompress import kexists
    # TODO glob
    # TODO not sure about windows path separators??
    spath = 'Takeout/My Activity/' + kind
    if not kexists(takeout, spath):
        logger.warning(f"{spath} is not present in {takeout}... skipping")
        return
    logger.info('processing %s %s', takeout, kind)

    locator = Loc.file(spath)
    from my.google.takeout.html import read_html
    for dt, url, title in read_html(takeout, spath):
        yield Visit(
            url=url,
            dt=dt,
            locator=locator,
            debug=kind,
        )

def _cpath(suffix: str):
    def fun(takeout: TakeoutPath):
        cache_dir = config.get().cache_dir
        if cache_dir is None:
            return None
        # doesn't need a nontrivial hash function, timestsamp is encoded in name
        return cache_dir / (takeout.name + '_' + suffix + '.cache')
    return fun


# todo caching should this be HPI responsibility?
# todo set global cachew logging on init?
@cachew(cache_path=_cpath('google_activity') , logger=logger)
def read_google_activity(takeout: TakeoutPath) -> Iterable[Visit]:
    return _read_myactivity_html(takeout, 'Chrome/MyActivity.html')

@cachew(cache_path=_cpath('search_activity') , logger=logger)
def read_search_activity(takeout: TakeoutPath) -> Iterable[Visit]:
    return _read_myactivity_html(takeout, 'Search/MyActivity.html')

# TODO add this to tests?
@cachew(cache_path=_cpath('browser_activity'), logger=logger)
def read_browser_history_json(takeout: TakeoutPath) -> Iterable[Visit]:
    from my.core.kompress import kexists, kopen
    # not sure if this deserves moving to HPI? it's pretty trivial for now
    spath = 'Takeout/Chrome/BrowserHistory.json'

    if not kexists(takeout, spath):
        logger.warning(f"{spath} is not present in {takeout}... skipping")
        return
    logger.info('processing %s %s', takeout, spath)

    # TODO couls also add spath?
    locator = Loc.file(takeout)

    # TODO this should be supported by HPI now?

    j = None
    with kopen(takeout, spath) as fo: # TODO iterative parser?
        j = json.load(fo)

    hist = j['Browser History']
    for item in hist:
        url = item['url']
        time = datetime.utcfromtimestamp(item['time_usec'] / 10 ** 6).replace(tzinfo=pytz.utc)
        # TODO any more interesitng info?
        yield Visit(
            url=url,
            dt=time,
            locator=locator,
            debug='Chrome/BrowserHistory.json',
        )




from contextlib import contextmanager
from datetime import datetime, timedelta
import json
import os
from pathlib import Path
from shutil import copy
import signal
import sys
from subprocess import check_output, check_call, PIPE
from textwrap import dedent
import time
from typing import NamedTuple, ContextManager, Optional

import pytz
import requests

from promnesia.common import PathIsh, _is_windows

from integration_test import index_hypothesis, index_urls, index_some_demo_visits
from common import tdir, under_ci, tdata, tmp_popen, promnesia_bin


class Helper(NamedTuple):
    port: str


# TODO use proper random port
TEST_PORT = 16556


def next_port():
    global TEST_PORT
    TEST_PORT += 1
    return TEST_PORT


@contextmanager
def wserver(db: Optional[PathIsh]=None): # TODO err not sure what type should it be... -> ContextManager[Helper]:
    port = str(next_port())
    cmd = [
        'serve',
        '--quiet',
        '--port', port,
        *([] if db is None else ['--db'  , str(db)]),
    ]
    with tmp_popen(promnesia_bin(*cmd)) as server:
        # wait till ready
        st = f'http://localhost:{port}/status'
        for a in range(50):
            try:
                requests.get(st).json()
                break
            except:
                time.sleep(0.1)
        else:
            raise RuntimeError("Cooldn't connect to '{st}' after 50 attempts")
        print("Started server up, db: {db}".format(db=db), file=sys.stderr)

        yield Helper(port=port)

        print("Done with the server", file=sys.stderr)


@contextmanager
def _test_helper(tmp_path):
    tdir = Path(tmp_path)
    cache_dir = tdir / 'cache'
    cache_dir.mkdir()

    # TODO extract that into index_takeout?
    # TODO ugh. quite hacky...
    template_config = tdata('test_config.py')
    copy(template_config, tdir)
    config = tdir / 'test_config.py'
    with config.open('a') as fo:
        fo.write(f"""
OUTPUT_DIR = r'{tdir}'
CACHE_DIR  = r'{cache_dir}'
""")

    check_call(promnesia_bin('index', '--config', config))

    with wserver(db=tdir / 'promnesia.sqlite') as srv:
        yield srv


def post(*args):
    cmd = [
        sys.executable, '-m', 'httpie',
        # '--timeout', '10000', # useful for debugging
        '--ignore-stdin',
        'post',
        *args,
    ]
    return json.loads(check_output(cmd).decode('utf8'))


def test_query_instapaper(tdir):
    index_hypothesis(tdir)
    test_url = "http://www.e-flux.com/journal/53/59883/the-black-stack/"
    with wserver(db=tdir / 'promnesia.sqlite') as helper:
        response = post(f'http://localhost:{helper.port}/visits', f'url={test_url}')
        assert len(response['visits']) > 5
        # TODO actually test response?


def test_visits(tmp_path: Path) -> None:
    test_url = 'https://takeout.google.com/settings/takeout'
    with _test_helper(tmp_path) as helper:
        # TODO why are we querying 3 times ???
        for q in range(3):
            print(f"querying {q}")
            response = post(f'http://localhost:{helper.port}/visits', f'url={test_url}')
            assert len(response['visits']) == 1


def test_search(tdir):
    index_hypothesis(tdir)
    test_url = "http://www.e-flux.com"
    with wserver(db=tdir / 'promnesia.sqlite') as helper:
        response = post(f'http://localhost:{helper.port}/search', f'url={test_url}')
        assert len(response['visits']) == 8


def test_visited(tmp_path: Path) -> None:
    test_url = 'https://takeout.google.com/settings/takeout'
    with _test_helper(tmp_path) as helper:
        endp = f'http://localhost:{helper.port}/visited'
        assert post(endp, f'''urls:=["{test_url}","http://badurl.org"]''') == [True, False]
        assert post(endp, f'''urls:=[]''') == []


import pytest
@pytest.mark.parametrize('count',
    [1, 5, 10, 12, 25, 50, 100, 200, 400, 800, 1600]
)
def test_visited_benchmark(count: int, tmp_path) -> None:
    pytest.skip("Only works on @karlicoss computer for now")
    # TODO skip on ci
    import promnesia.server as S
    # TODO reset after?
    S.EnvConfig.set(S.ServerConfig(
        # TODO populate with test db and benchmark properly...
        db=Path('/todo'),
        timezone=pytz.utc,
    ))
    links = [f'https://reddit.com/whatever{i}.html' for i in range(count)]
    res = S.visited(links)
    assert len(res) == len(links)


def test_search_around(tmp_path: Path) -> None:
    # EDT, should be UTC-4
    dt_extra = pytz.timezone('America/New_York').localize(datetime.fromisoformat('2018-06-01T10:00:00.000000'))
    # NOTE: negative timedelta to test that it captures some past context
    # at the moment it's hardcoded by delta_back/delta_front -- would be nice to make it configurable later...
    index_some_demo_visits(tmp_path, count=1000, base_dt=dt_extra, delta=-timedelta(minutes=1), update=False)

    # TODO hmm. perhaps it makes more sense to run query in different process and server in main process for testing??
    with wserver(db=tmp_path / 'promnesia.sqlite') as helper:
        response = post(f'http://localhost:{helper.port}/search_around', f'timestamp={int(dt_extra.timestamp())}')
        assert len(response['visits']) > 10, response


# TODO right.. I guess that triggered because of reddit indexer specifically
# TODO could probably reuse parts of tests to be both integration/server and end2end when necessary?
def test_visits_hier(tdir):
    test_url = 'https://www.reddit.com/r/QuantifiedSelf/comments/d6m7bd/android_app_to_track_and_export_application_use/'
    urls = {
        test_url: 'parent url',
        'https://reddit.com/r/QuantifiedSelf/comments/d6m7bd/android_app_to_track_and_export_application_use/f0vem56': 'Some context',
        'https://reddit.com/r/QuantifiedSelf/comments/d6m7bd/android_app_to_track_and_export_application_use/whatever': None, # no context so should be ignored..
    }
    indexer = index_urls(urls)
    indexer(tdir)
    with wserver(db=tdir / 'promnesia.sqlite') as helper:
        response = post(f'http://localhost:{helper.port}/visits', f'url={test_url}')
        assert {v['context'] for v in response['visits']} == {'parent url', 'Some context'}


def test_status(tdir):
    with wserver(db='/does/not/exist') as helper:
        response = post(f'http://localhost:{helper.port}/status')
        assert 'ERROR' in response['db'] # defensive, it doesn't exist
        version = response['version']
        assert version is not None
        assert len(version.split('.')) >= 2 # random check..


def test_basic(tmp_path: Path) -> None:
    cfg = tmp_path / 'config.py'
    cfg.write_text("SOURCES = ['promnesia.sources.demo']")
    check_call(promnesia_bin('index', '--config', cfg))
    with wserver() as helper:
        response = post(f'http://localhost:{helper.port}/visits', 'url=whatever')
        assert response['visits'] == []


def test_query_while_indexing(tmp_path: Path) -> None:
    cfg = tmp_path / 'config.py'
    indexing_cmd = promnesia_bin('index', '--config', cfg)

    # just trigger the database
    cfg.write_text(dedent(f'''
    OUTPUT_DIR = r'{tmp_path}'
    SOURCES = ['promnesia.sources.demo']
    '''))
    check_call(indexing_cmd)

    cfg.write_text(dedent(f'''
    OUTPUT_DIR = r'{tmp_path}'

    from promnesia.common import Source
    from promnesia.sources import demo
    # index stupid amount of visits to increase time spent in database serialization
    SOURCES = [Source(demo.index, count=100000)]
    '''))
    with wserver(db=tmp_path / 'promnesia.sqlite') as helper:
        status = lambda: post(f'http://localhost:{helper.port}/status')
        # precondition -- db should be healthy
        r = status()
        assert 0 < r['stats']['total_visits'] < 100000, r

        # now run the indexing (asynchronously)
        #
        from subprocess import Popen
        with Popen(indexing_cmd):
            # and hammer the backend to increase likelihood of race condition
            # not ideal -- doesn't really 'guarantee' to catch races, but good enough
            for _ in range(100):
                r = status()
                assert r['stats'].get('total_visits', 0) > 0, r
        # after indexing finished, new visits should be in the db
        r = status()
        assert r['stats']['total_visits'] >= 100000, r



from pathlib import Path
from promnesia.common import traverse
from unittest.mock import Mock, patch
from common import DATA


testDataPath = Path(DATA) / 'traverse'

# Patch shutil.which so it always returns false (when trying to which fdfind, etc)
# so that it falls back to find
@patch('promnesia.common.shutil.which', return_value=False)
def test_traverse_ignore_find(patched):
    '''
    traverse() with `find` but ignore some stuff
    '''
    # act
    paths = set(traverse(testDataPath, ignore=['ignoreme.txt', 'ignoreme2']))

    # assert
    assert paths == {testDataPath / 'imhere2/real.txt', testDataPath / 'imhere.txt'}

def test_traverse_ignore_fdfind():
    '''
    traverse() with `fdfind` but ignore some stuff
    '''
    # act
    paths = set(traverse(testDataPath, ignore=['ignoreme.txt', 'ignoreme2']))

    # assert
    assert paths == {testDataPath / 'imhere.txt', testDataPath / 'imhere2/real.txt'}

# TODO: It would be nice to test the implementation directly without having to do this
# weird patching in the future
@patch('promnesia.common._is_windows', new_callable=lambda: True)
def test_traverse_ignore_windows(patched):
    '''
    traverse() with python when _is_windows is true but ignore some stuff
    '''
    # act
    paths = set(traverse(testDataPath, ignore=['ignoreme.txt', 'ignoreme2']))

    # assert
    assert paths == {testDataPath / 'imhere.txt', testDataPath / 'imhere2/real.txt'}



"""
Adapted from `telegram.py` to read from `~/.ViberPC/XYZ123/viber.db`
"""

import logging
import textwrap
from os import PathLike
from pathlib import Path
from typing import Iterable, Optional

from ..common import Loc, PathIsh, Results, Visit, extract_urls, from_epoch, join_tags

logger = logging.getLogger(__name__)


def index(
    db_path: PathIsh = "~/.ViberPC/*/viber.db",
    locator_schema="editor",
    *,
    http_only: bool = None,
) -> Results:
    """
    :param db_path:
        the path of the sqlite maintained by the VibdePC desktop app
    :param http_only:
        when true, do not collect IP-addresses and `python.py` strings
    """
    glob_paths = list(_get_files(db_path))
    logger.debug("Expanded path(s): %s", glob_paths)
    assert glob_paths, f"No Viber-desktop sqlite found: {db_path}"

    msgs_query = messages_query(http_only)

    for db_path in _get_files(db_path):
        assert db_path.is_file(), f"Is it a (Viber-desktop sqlite) file? {db_path}"
        yield from _harvest_db(db_path, msgs_query, locator_schema)


# TODO move to common?
def _dataset_readonly(db: Path):
    # see https://github.com/pudo/dataset/issues/136#issuecomment-128693122
    import sqlite3

    import dataset  # type: ignore

    creator = lambda: sqlite3.connect(f"file:{db}?immutable=1", uri=True)
    return dataset.connect("sqlite:///", engine_kwargs={"creator": creator})


def messages_query(http_only: Optional[bool]) -> str:
    """
    An SQL-query returning 1 row for each message

    A non-private method, to facilitate experimentation.
    """
    extra_criteria = "AND text LIKE '%http%'" if http_only else ""
    return textwrap.dedent(
        f"""
        /*
        Establish group-names by concatenating:
        - groups explicitly named,
        - multi-groups having a group-leader (PGRole=2), with
        - all the rest groups that (must) have just 2 members,
          me(ContactId=1) & "other+" contacts, so use "other"" as group-name.
        */
        WITH G0 AS (
            SELECT
                CR.ChatId,
                CR.ContactID,
                coalesce(CI.Name, C.Name, C.ClientName) as chatname,
                CR.PGRole,
                CI.PGTags
            FROM ChatRelation as CR
            JOIN Contact AS C ON CR.ContactID = C.ContactID
            JOIN ChatInfo as CI ON CR.ChatId = CI.ChatId
        ), G1 AS (
            SELECT * FROM G0 WHERE PGRole = 2
        ), G2 AS (
            SELECT * FROM G0 WHERE
                ContactID <> 1 AND
                ChatId NOT IN (SELECT ChatId FROM G1)
        ), Groups AS (
            SELECT ChatId, chatname, PGTags FROM G1
            UNION
            SELECT ChatId, chatname, PGTags FROM G2
        )
        SELECT
            M.EventId           AS mid,
            E.TimeStamp         AS time,
            G.chatname          AS chatname,
            coalesce(
                S.Name,
                S.ClientName,
                '(' || S.Number || ')'          /* contacts have one xor the other, but failsafe */
            )                   AS sender,
            coalesce(M.Subject, M.Body)         /* didn't see any msg with both */
                                AS text,
            IIF(
                json_valid(M.info),
                json_extract(M.info, '$.Title'),
                ''
            ) as url_title,                     /* harvested titles from embedded urls */
            G.PGTags            AS tags
        FROM messages AS M
        LEFT JOIN Events AS E
            ON M.EventId = E.EventId
        LEFT JOIN Contact AS S
            ON E.ContactId = S.ContactId
        LEFT JOIN Groups AS G
            ON E.ChatId = G.ChatId
        WHERE
            M.ClientFlag != 64                  /* edited messages */
            AND text IS NOT NULL
            {extra_criteria}
        ORDER BY time;
        """
    )


def _handle_row(row: dict, db_path: PathLike, locator_schema: str) -> Results:
    text = row["text"]
    urls = extract_urls(text)
    if not urls:
        return

    dt = from_epoch(row["time"] // 1000)  # timestamps are stored x100 this db
    mid: str = row["mid"]
    # TODO perhaps we could be defensive with null sender/chat etc and still emit the Visit
    sender: str = row["sender"]
    chatname: str = row["chatname"]
    tags: str = row["tags"]
    url_title: str = row["url_title"]

    assert (
        text and mid and sender and chatname
    ), f"sql-query should eliminate messages without 'http' or missing ids: {row}"

    if tags and tags.strip():
        tags = join_tags(tags.split())
        text = f"{text}\n\n{tags}"

    if url_title:
        text = f"title: {url_title}\n\n{text}"

    for u in urls:
        yield Visit(
            url=u,  # URLs in Viber's SQLite are not quoted
            dt=dt,
            context=text,
            locator=Loc.make(
                title=f"chat({mid}) from {sender}@{chatname}",
                href=f"{locator_schema}://{db_path}#!Messages.EventId={mid}",
            ),
        )


def _get_files(path: PathIsh) -> Iterable[Path]:
    """
    Expand homedir(`~`) and return glob paths matched.

    Expansion code copied from https://stackoverflow.com/a/51108375/548792
    """
    path = Path(path).expanduser()
    parts = path.parts[1:] if path.is_absolute() else path.parts
    return Path(path.root).glob(str(Path("").joinpath(*parts)))


def _harvest_db(db_path: PathIsh, msgs_query: str, locator_schema: str):
    is_debug = logger.isEnabledFor(logging.DEBUG)

    # Note: for displaying maybe better not to expand/absolute,
    # but it's safer for debugging resolved.
    db_path = Path(db_path).resolve()

    with _dataset_readonly(db_path) as db:
        for row in db.query(msgs_query):
            try:
                yield from _handle_row(row, db_path, locator_schema)
            except Exception as ex:
                # TODO: also insert errors in db
                logger.warning(
                    "Cannot extract row: %s, due to: %s(%s)",
                    row,
                    type(ex).__name__,
                    ex,
                    exc_info=is_debug,
                )



#!/usr/bin/env python3
from functools import lru_cache
from pathlib import Path
from typing import Dict, Callable, Optional, Sequence, NamedTuple, Union, Iterable

from ..common import Results, Url


# TODO doesn't really belong here...
Ctx = Sequence[str]

class EUrl(NamedTuple):
    url: Url
    ctx: Ctx # TODO ctx here is more like a Loc
###


# keys are mime types + extensions
Ex = Callable[[Path], Union[Results, Iterable[EUrl]]]
# None means unhandled
TYPE2IDX: Dict[str, Optional[Ex]] = {}
# NOTE: there are some types in auto.py at the moment... it's a bit messy


# TYPE2IDX only contains the 'prefixes', to speed up the lookup we are using cache..
@lru_cache(None)
def type2idx(t: str) -> Optional[Ex]:
    if len(t) == 0:
        return None # just in case?
    # first try exact match
    e = TYPE2IDX.get(t, None)
    if e is not None:
        return e
    t = t.strip('.')
    e = TYPE2IDX.get(t, None)
    if e is not None:
        return e
    # otherwise, try prefixes?
    for k, v in TYPE2IDX.items():
        if t.strip('.').startswith(k):
            return v
    return None

# for now source code just indexed with grep, not sure if it's good enough?
# if not, some fanceir library could be used...
# e.g. https://github.com/karlicoss/promnesia/pull/152/commits/c2f00eb4ee4018b02c9bf3966a036db69a43373d

# TODO use this list?
# https://github.com/GerritCodeReview/gerrit/blob/master/resources/com/google/gerrit/server/mime/mime-types.properties
# later these might do something clever, e.g. stripping off code comments etc?
CODE = {
    'text/x-java',
    'text/x-tex',
    'text/x-sh',
    'text/x-haskell',
    'text/x-perl',
    'text/x-python', 'text/x-script.python',
    'text/x-chdr',
    'text/x-csrc',
    'text/x-c',
    'text/x-c++',
    'text/x-makefile',
    'text/troff',
    'text/x-asm',
    'text/x-objective-c',
    'text/x-lisp',
    'text/vnd.graphviz',
    'text/x-diff',  # patch files

    # these didn't have a mime type, or were mistyped?
    'css',
    'el',
    'rs',
    'go',
    'hs',  # mistyped on osx
    'hpp', # mistyped on osx

    'edn', # clojure data

    '.ts', # most likely typescript.. otherwise determined as text/vnd.trolltech.linguist mime
}
# TODO discover more extensions with mimetypes library?


BINARY = '''
# epub was failing to detect via mime on CI for some reason..
epub
inode/x-empty
.sqlite
# comment
application/
image/
audio/
video/
'''

handle_later = lambda *args, **kwargs: ()

def ignore(*args, **kwargs):
    # TODO log (once?)
    yield from ()


for x in BINARY.splitlines():
    x = x.strip()
    if len(x) == 0 or x[0] == '#':
        continue
    TYPE2IDX[x] = ignore


TYPE2IDX.update({
    '.xslx': ignore,
    '.vcf' : ignore,
    'message/rfc822': ignore, # ??

    # TODO not sure what to do about these..
    'application/octet-stream': handle_later,
    'application/zip'         : handle_later,
    'application/x-tar'       : handle_later,
    'application/gzip'        : handle_later,
})


# TODO use some existing file for initial gitignore..
IGNORE = [
    '.idea',
    'venv',
    '.git',
    '.eggs',
    '.mypy_cache',
    '.pytest_cache',
    'node_modules',
    '__pycache__',
    '.tox',
    '.stack-work',

    # TODO not sure about these:
    '.gitignore',
    '.babelrc',
]




from datetime import datetime
import os
import re
from typing import Optional, Union
import warnings

from ..compat import run, PIPE, Paths
from ..common import Visit, Loc, Results, extract_urls, file_mtime, get_system_tz, now_tz, _is_windows
from .plaintext import _has_grep


def index(command: Union[str, Paths]) -> Results:
    cmd: Paths
    cmds: str
    if isinstance(command, str):
        cmds = command
        warnings.warn("Passing string as a command is very fragile('{command}'). Please use list instead.")
        cmd = command.split(' ')
    else:
        cmds = ' '.join(map(str, command))
        cmd = command

    tz = get_system_tz()

    # ugh... on windows grep does something nasty? e.g:
    # grep --color=never -r -H -n -I -E http 'D:\\a\\promnesia\\promnesia\\tests\\testdata\\custom'
    # D:\a\promnesia\promnesia\tests\testdata\custom/file1.txt:1:Right, so this points at http://google.com
    # so part of the path has fwd slashes, part has bwd slashes...
    needs_windows_grep_patching = _has_grep() and _is_windows

    def handle_line(line: str) -> Results:
        # grep dumps this as
        # /path/to/file:lineno:rest
        # note: on Windows, path contains : after the disk name..
        m = re.search(r'(.*?):(\d+?):(.*)', line)
        if m is None:
            # todo warn maybe?
            fname = None
            lineno = None
        else:
            fname  = m.group(1)
            lineno = int(m.group(2))
            line   = m.group(3)

        if fname is not None and needs_windows_grep_patching:
            fname = fname.replace('/', os.sep)

        urls = extract_urls(line)
        if len(urls) == 0:
            return

        context = line

        ts: datetime
        loc: Loc
        if fname is not None:
            ts = file_mtime(fname)
            loc = Loc.file(fname, line=lineno)
        else:
            ts = now_tz()
            loc = Loc.make(cmds)
        for url in urls:
            yield Visit(
                url=url,
                dt=ts,
                locator=loc,
                context=context,
            )

    r = run(cmd, stdout=PIPE)
    if r.returncode > 0:
        if not (cmd[0] in {'grep', 'findstr'} and r.returncode == 1): # ugh. grep returns 1 on no matches...
            r.check_returncode()
    output = r.stdout
    assert output is not None
    lines = [line.decode('utf-8') for line in output.splitlines()]
    for line in lines:
        try:
            yield from handle_line(line)
        except Exception as e:
            yield e



'''
Uses HPI [[https://github.com/karlicoss/HPI/blob/master/doc/MODULES.org#mygoogletakeoutpaths][google.takeout]] module
'''
from typing import Iterable, Set, Type
import warnings

from ..common import Visit, Loc, Results, logger
from ..compat import removeprefix


def index() -> Results:
    from . import hpi

    try:
        from my.google.takeout.parser import events
        from google_takeout_parser.models import Activity, YoutubeComment, LikedYoutubeVideo, ChromeHistory, PlayStoreAppInstall, Location
    except ModuleNotFoundError as ex:
        logger.exception(ex)
        yield ex

        warnings.warn("Please set up my.google.takeout.parser module for better takeout support. Falling back to legacy implementation.")

        from . import takeout_legacy
        yield from takeout_legacy.index()
        return

    _seen: Set[Type] = {
        # these are definitely not useful for promnesia
        Location,
        PlayStoreAppInstall,
    }
    def warn_once_if_not_seen(e) -> Iterable[Exception]:
        et = type(e)
        if et in _seen:
            return
        _seen.add(et)
        yield RuntimeError(f"Unhandled event {et}: {e}")

    for e in events():
        if isinstance(e, Exception):
            yield e
            continue
        elif isinstance(e, Activity):
            # TODO: regex out title and use it as locator title?
            url = e.titleUrl
            if url is not None:
                # when you follow something from search the actual url goes after this
                # e.g. https://www.google.com/url?q=https://en.wikipedia.org/wiki/Clapham
                # note: also title usually starts with 'Visited ', in such case but perhaps fine to keep it
                url = removeprefix(url, "https://www.google.com/url?q=")

                yield Visit(
                    url=url,
                    dt=e.time,
                    context=e.header,
                    locator=Loc(title=e.title, href=url),
                )
            for s in e.subtitles:
                surl = s[1]
                if surl is not None:
                    if "youtube.com/channel" in surl:
                        continue
                    yield Visit(
                        url=surl,
                        dt=e.time,
                        context=s[0],
                        locator=Loc(title=e.title, href=surl),
                    )
        elif isinstance(e, ChromeHistory):
            yield Visit(
                url=e.url,
                dt=e.dt,
                locator=Loc(title=e.title, href=e.url),
            )
        elif isinstance(e, LikedYoutubeVideo):
            yield Visit(
                url=e.link, dt=e.dt, context=e.desc, locator=Loc(title=e.title, href=e.link)
            )
        elif isinstance(e, YoutubeComment):
            for url in e.urls:
                # todo: use url_metadata to improve locator?
                # or maybe just extract first sentence?
                yield Visit(
                    url=url, dt=e.dt, context=e.content, locator=Loc(title=e.content, href=url)
                )
        else:
            yield from warn_once_if_not_seen(e)



from pathlib import Path
import shutil
from typing import Dict, List, Tuple, Set, Iterable

from more_itertools import chunked

from sqlalchemy import create_engine, MetaData, event # type: ignore
from sqlalchemy import Column, Table # type: ignore

from cachew import NTBinder

from .common import get_logger, DbVisit, get_tmpdir, Res, now_tz, Loc
from . import config


# NOTE: I guess the main performance benefit from this is not creating too many tmp lists and avoiding overhead
# since as far as sql is concerned it should all be in the same transaction. only a guess
# not sure it's the proper way to handle it
# see test_index_many
_CHUNK_BY = 10

# I guess 1 hour is definitely enough
_CONNECTION_TIMEOUT_SECONDS = 3600

# returns critical warnings
def visits_to_sqlite(vit: Iterable[Res[DbVisit]], *, overwrite_db: bool) -> List[Exception]:
    logger = get_logger()
    db_path = config.get().db

    now = now_tz()
    ok = 0
    errors = 0
    def vit_ok() -> Iterable[DbVisit]:
        nonlocal errors, ok
        for v in vit:
            if isinstance(v, DbVisit):
                ok += 1
                yield v
            else:
                errors += 1
                # conform to the schema and dump. can't hurt anyway
                ev = DbVisit(
                    norm_url='<error>',
                    orig_url='<error>',
                    dt=now,
                    locator=Loc.make('<errror>'),
                    src='error',
                    # todo attach backtrace?
                    context=repr(v),
                )
                yield ev

    tpath = Path(get_tmpdir().name) / 'promnesia.tmp.sqlite'
    if overwrite_db:
        # here we don't need timeout, since it's a brand new DB
        engine = create_engine(f'sqlite:///{tpath}')
    else:
        # here we need a timeout, othewise concurrent indexing might not work
        # (note that this also needs WAL mode)
        # see test_concurrent_indexing
        engine = create_engine(f'sqlite:///{db_path}', connect_args={'timeout': _CONNECTION_TIMEOUT_SECONDS})

    # using WAL keeps database readable while we're writing in it
    # this is tested by test_query_while_indexing
    def enable_wal(dbapi_con, con_record):
        dbapi_con.execute('PRAGMA journal_mode = WAL')
    event.listen(engine, 'connect', enable_wal)

    binder = NTBinder.make(DbVisit)
    meta = MetaData(engine)
    table = Table('visits', meta, *binder.columns)
    meta.create_all()

    cleared: Set[str] = set()
    ncleared = 0
    with engine.begin() as conn:
        for chunk in chunked(vit_ok(), n=_CHUNK_BY):
            srcs = set(v.src or '' for v in chunk)
            new = srcs.difference(cleared)

            for src in new:
                conn.execute(table.delete().where(table.c.src == src))
                ncleared += conn.execute("SELECT changes()").fetchone()[0]
                cleared.add(src)

            bound = [binder.to_row(x) for x in chunk]
            # pylint: disable=no-value-for-parameter
            conn.execute(table.insert().values(bound))

    if overwrite_db:
        shutil.move(str(tpath), str(db_path))

    errs = '' if errors == 0 else f', {errors} ERRORS'
    total = ok + errors
    what = 'overwritten' if overwrite_db else 'updated'
    logger.info(
        '%s database "%s". %d total (%d OK%s, %d cleared, +%d more)',
        what, db_path, total, ok, errs, ncleared, ok - ncleared)
    res: List[Exception] = []
    if total == 0:
        res.append(RuntimeError('No visits were indexed, something is probably wrong!'))
    return res



import os.path
import urllib.parse

def logseq_replacer(path: str, root: str) -> str:
    if not path.startswith("editor://") or not (path.endswith('.md') or path.endswith('.org')):
        return path
        
    graph = os.path.basename(root)
    page_name = os.path.basename(path).rsplit('.', 1)[0]
    encoded_page_name = urllib.parse.quote(page_name)
    
    uri = f"logseq://graph/{graph}?page={encoded_page_name}"

    return uri



"""
- discovers files recursively
- guesses the format (orgmode/markdown/json/etc) by the extension/MIME type
"""

import csv
from concurrent.futures import ProcessPoolExecutor as Pool
from contextlib import nullcontext
from datetime import datetime
import itertools
import json
import os
from typing import Optional, Iterable, Union, List, Tuple, NamedTuple, Sequence, Iterator, Iterable, Callable, Any, Dict, Set
from fnmatch import fnmatch
from pathlib import Path
from functools import lru_cache, wraps
import warnings

import pytz

from ..common import Visit, Url, PathIsh, get_logger, Loc, get_tmpdir, extract_urls, Extraction, Result, Results, mime, traverse, file_mtime, echain, logger
from ..config import use_cores


from .filetypes import EUrl
from .auto_obsidian import obsidian_replacer
from .auto_logseq import logseq_replacer


def _collect(thing, path: List[str], result: List[EUrl]) -> None:
    if isinstance(thing, str):
        ctx: Ctx = tuple(path) # type: ignore
        result.extend([EUrl(url=u, ctx=ctx) for u in extract_urls(thing)])
    elif isinstance(thing, list):
        path.append('[]')
        for x in thing:
            _collect(x, path, result)
        path.pop()
    elif isinstance(thing, dict):
        for k, v in thing.items():
            path.append(k)
            _collect(k, path, result)
            _collect(v, path, result)
            path.pop()
    else:
        pass


# TODO mm. okay, I suppose could use kython consuming thingy?..
def collect_from(thing) -> List[EUrl]:
    uuu: List[EUrl] = []
    path: List[str] = []
    _collect(thing, path, uuu)
    return uuu


Urls = Iterator[EUrl]

def _csv(path: Path) -> Urls:
    # TODO these could also have Loc to be fair..
    with path.open() as fo:
        # TODO shit need to urldecode
        reader = csv.DictReader(fo)
        for line in reader:
            yield from collect_from(line)


def _json(path: Path) -> Urls:
    jj = json.loads(path.read_text())
    yield from collect_from(jj)


def _plaintext(path: Path) -> Results:
    from . import shellcmd
    from .plaintext import extract_from_path
    yield from shellcmd.index(extract_from_path(path))


# TODO think about the type
# TODO could pass fallback reason to the results as well?
def fallback(ex):
    """Falls back to plaintext in case of issues"""

    fallback_active: Dict[Any, bool] = {}
    @wraps(ex)
    def wrapped(path: Path):
        nonlocal fallback_active
        do_fallback = fallback_active.get(ex, False)

        if not do_fallback:
            try:
                it = ex(path)
                # ugh. keeping yield in the try section is not ideal, but need this because of lazy yield semantics
                yield from it
            except ModuleNotFoundError as me:
                logger = get_logger()
                logger.exception(me)
                logger.warn('%s: %s not found, falling back to grep! "pip3 install --user %s" for better support!', path, me.name, me.name)
                yield me
                fallback_active[ex] = True
                do_fallback = True
        if do_fallback:
            yield from _plaintext(path)
    return wrapped


@fallback
def _markdown(path: Path) -> Results:
    from . import markdown
    yield from markdown.extract_from_file(path)


@fallback
def _html(path: Path) -> Results:
    from . import html
    yield from html.extract_from_file(path)


@fallback
def _org(path: Path) -> Results:
    from . import org
    return org.extract_from_file(path)


from .filetypes import TYPE2IDX, type2idx, IGNORE, CODE

TYPE2IDX.update({
    'application/json': _json,
    '.json'           : _json,
    '.ipynb'          : _json,

    '.csv'           : _csv,
    'application/csv': _csv,

    '.org'        : _org,
    '.org_archive': _org,

    '.md'         : _markdown,
    '.markdown'   : _markdown,

    'text/plain'  : _plaintext,
    '.txt'        : _plaintext,
    '.page'       : _plaintext,
    '.rst'        : _plaintext,


    # TODO doesn't work that great; weird stuff like
    # builtins.ImportError.name|2019-07-10T12:12:35.584510+00:00|names::ImportError::node::names::name::node::fullname
    # TODO could have stricter url extraction for that; always using http/https?
    # '.ipynb'      : _json,

    '.html'    : _html,
    'text/html': _html,
    'text/xml' : _plaintext,

    'text/x-po': _plaintext, # some translation files
})

for t in CODE:
    TYPE2IDX[t] = _plaintext
# TODO ok, mime doesn't really tell between org/markdown/etc anyway


Replacer = Optional[Callable[[str, str], str]]

def index(
        *paths: Union[PathIsh],
        ignored: Union[Sequence[str], str]=(),
        follow: bool=True,
        replacer: Replacer=None,
) -> Results:
    '''
    path   : a path or list of paths to recursively index
    ignored: a glob or list of globs to exclude from indexing
    follow : whether to follow symlinks or not
    '''
    # TODO document replacer?
    ignored = (ignored,) if isinstance(ignored, str) else ignored
    for p in paths:
        # TODO for displaying maybe better not to expand/absolute, but need it for correct mime handling
        apath = Path(p).expanduser().resolve().absolute()
        root = apath if apath.is_dir() else None

        if root is not None and (root / ".obsidian").exists():
            if replacer:
                logger.debug('detected %s as Obsidian vault, but not changing replacer', root)
            else:
                logger.debug('detected %s as Obsidian vault', root)
                replacer = obsidian_replacer

        if root is not None and (root / "logseq/config.edn").exists():
            logger.debug('detected %s as Logseq graph', root)
            replacer = logseq_replacer

        opts = Options(
            ignored=ignored,
            follow=follow,
            replacer=replacer,
            root=root,
        )
        yield from _index(apath, opts=opts)

class Options(NamedTuple):
    ignored: Sequence[str]
    follow: bool
    # TODO option to add ignores? not sure..
    # TODO I don't like this replacer thing... think about removing it
    replacer: Replacer
    root: Optional[Path]=None


def _index_file_aux(path: Path, opts: Options) -> Union[Exception, List[Result]]:
    # just a helper for the concurrent version (the generator isn't picklable)
    try:
        return list(_index_file(path, opts=opts))
    except Exception as e:
        # possible due to unavoidable race conditions
        return e


def _index(path: Path, opts: Options) -> Results:
    logger = get_logger()

    cores = use_cores()
    if cores is None: # do not use cores
        # todo use ExitStack instead?
        pool = nullcontext()
        mapper = map # dummy pool
    else:
        workers = None if cores == 0 else cores
        pool = Pool(workers) # type: ignore
        mapper = pool.map # type: ignore

    # iterate over resolved paths, to avoid duplicates
    def rit() -> Iterable[Path]:
        it = traverse(path, follow=opts.follow, ignore=IGNORE)
        for p in it:
            if any(fnmatch(str(p), o) for o in opts.ignored):
                # TODO not sure if should log here... might end up with quite a bit of logs
                logger.debug('ignoring %s: user ignore rules', p)
                continue
            if any(i in p.parts for i in IGNORE): # meh, not very efficient.. pass to traverse??
                logger.debug('ignoring %s: default ignore rules', p)
                continue

            p = p.resolve()
            if not os.path.exists(p):
                logger.debug('ignoring %s: broken symlink?', p)
                continue

            yield p

    from more_itertools import unique_everseen
    it = unique_everseen(rit())

    with pool:
        for r in mapper(_index_file_aux, it, itertools.repeat(opts)):
            if isinstance(r, Exception):
                yield r
            else:
                yield from r


Mime = str
from .filetypes import Ex # meh
def by_path(pp: Path) -> Tuple[Optional[Ex], Optional[Mime]]:
    suf = pp.suffix.lower()
    # firt check suffixes, it's faster
    s = type2idx(suf)
    if s is not None:
        return s, None
    # then try with mime
    pm = mime(pp)
    if pm is not None:
        return type2idx(pm), pm
    else:
        return None, None


def _index_file(pp: Path, opts: Options) -> Results:
    logger = get_logger()
    # TODO use kompress?
    # TODO not even sure if it's used...
    suf = pp.suffix.lower()

    if suf == '.xz': # TODO zstd?
        import lzma
        uname = pp.name[:-len('.xz')]  # chop off suffix, so the downstream indexer can handle it

        assert pp.is_absolute(), pp
        # make sure to keep hierarchy, otherwise might end up with some name conflicts if filenames clash
        uncomp = Path(get_tmpdir().name) / Path(*pp.parts[1:-1]) / uname
        uncomp.parent.mkdir(parents=True, exist_ok=True)
        # todo would dump_file = wdir / Path(*cleaned_db.parts[1:])  # cut off '/' and use relative path
        with lzma.open(pp, 'rb') as cf:
            with uncomp.open('wb') as fb:
                fb.write(cf.read())
        # TODO maybe keep the original name?
        # currently it would end up with something like /tmp/tmpxpgx1jy2promnesia/reddit-20190401231025.json
        yield from _index(path=uncomp, opts=opts)
        return

    ex = RuntimeError(f'While indexing {pp}')

    ip, pm = by_path(pp)
    if ip is None:
        # TODO use warning (with mime/ext as key?)
        # TODO only log once? # hmm..
        msg = f'No extractor for suffix {suf}, mime {pm}'
        warnings.warn(msg)
        yield echain(ex, RuntimeError(msg))
        return

    logger.debug('indexing via %s: %s', ip.__name__, pp)

    def indexer() -> Union[Urls, Results]:
        # eh, annoying.. need to make more generic..
        idx = ip(pp) # type: ignore
        try:
            yield from idx
        except Exception as e:
            yield e

    root = opts.root
    fallback_dt = file_mtime(pp)
    fallback_loc = Loc.file(pp)
    replacer = opts.replacer
    for r in indexer():
        if isinstance(r, Exception):
            # indexers can rely on this method setting the error context
            yield echain(ex, r)
            continue
        if isinstance(r, EUrl):
            v = Visit(
                url=r.url,
                dt=fallback_dt,
                locator=fallback_loc,
                context='::'.join(r.ctx),
            )
        else:
            v = r

        loc = v.locator
        if loc is not None and root is not None:
            # meh. but it works
            # todo potentially, just use dataclasses instead...
            loc = loc._replace(title=loc.title.replace(str(root) + os.sep, ''))
            v = v._replace(locator=loc)

        if replacer is not None and root is not None:
            upd: Dict[str, Any] = {}
            href = v.locator.href
            if href is not None:
                upd['locator'] = v.locator._replace(href=replacer(href, str(root)), title=replacer(v.locator.title, str(root)))
            ctx = v.context
            if ctx is not None:
                # TODO in context, http is unnecessary
                upd['context'] = replacer(ctx, str(root))
            v = v._replace(**upd)
        yield v



'''
A more sophisticated example of config for Promnesia
'''

from promnesia.common import Source


# now, let's import some indexers
# NOTE: you might need extra dependencies before using some of the indexers
# see https://github.com/karlicoss/HPI/blob/master/doc/SOURCES.org

# 'auto' indexer tries its best at indexing plaintext stuff
# - plaintext like org-mode/markdown/HTML
# - structured formats like JSON and CSV
from promnesia.sources import auto

# 'guess' indexer can do even more in addition:
# - HTTP links (to index the contents of a website)
# - Github links (to index the contents of a git repository
from promnesia.sources import guess
# TODO there is a very thin link between 'auto' and 'guess'... I might merge them in the future?


# this is an incomplete list, just the (perhaps) most interesting ones
from promnesia.sources import telegram
from promnesia.sources import (
    fbmessenger,
    hypothesis,
    instapaper,
    pocket,
    roamresearch,
    rss,
    takeout,
    twitter,
    viber,
    signal,
)


# NOTE: at the moment try to avoid using complex sources names
# it's best to stick to digits, latin characters, dashes and underscores
# there might be some UI problems related to that

# now we can specify the sources
# this is a required setting
SOURCES = [
    # handle my knowledge base: extract links from Org-mode and Markdown files
    Source(
        auto.index,
        # NOTE: you'd need to specify your own filesystem path here
        '/path/to/my/blog',

        # you can specify optional name, so you can see where the URL is coming from within the extension
        name='blog',
    ),

    # we can index another path under a different name (for convenience)
    Source(
        auto.index,
        # NOTE: you'd need to specify your own filesystem path here
        '/path/to/personal/notes',
        ignored=['*.html'],  # we can exclude certain files if we want

        name='notes',
    ),

    # we can index Logseq graph: extract links from Org-mode and Markdown files
    Source(
        auto.index,
        '/path/to/logseq/graph',
        ignored=['*/.obsidian/*', '*/logseq/*', '*/assets/*'],
        name='logseq',
    ),

    # will clone the repository and index its contents!
    Source(
        guess.index, # you could also use 'vcs.index', but auto can handle it anyway
        'https://github.com/karlicoss/exobrain',
        name='exobrain',
    ),


    # Uses the output of telegram_backup tool: https://github.com/fabianonline/telegram_backup#usage
    # name will be set to 'telegram' by default
    Source(
        telegram,
        '/data/telegram/database.sqlite',
        #http_only=None  # harvest alsoIP-addresses & plain domain-names
    ),

    # Uses all local SQLite files found in your Viber Desktop configurations
    # (one directory for each telephone number):
    #     ~/.ViberPC/**/viber.db
    #
    # You may modify that by providing a 2nd ``Source()`` argument.
    Source(
        viber,
        #"path/to/viber-sql",
        #http_only=None  # harvest alsoIP-addresses & plain domain-names
    ),

    # When path(s) given, uses the SQLite inside Signal-Desktop's configuration directory
    # (see the sources for more parameters & location of the db-file for each platform)
    signal,

    # NOTE: to configure the following modules you need to set up HPI package (https://github.com/karlicoss/HPI#whats-inside)
    # see HPI setup guide    : https://github.com/karlicoss/HPI/blob/master/doc/SETUP.org
    # and HPI usage examples': https://github.com/karlicoss/HPI/blob/master/doc/SETUP.org#usage-examples
    ####
    Source(hypothesis.index),
    Source(takeout.index),

    # you can also be less verbose in config, if you prefer
    Source(roamresearch),
    Source(rss, name='custom name'),

    # or:
    fbmessenger.index,

    # or even just:
    pocket,
    instapaper,


    # sometimes lambdas are useful for config hacking/avoiding early imports
    lambda: twitter.index(),
]



''''
Optional setting.
A directory where promnesia.sqlite will be stored.
If not specified, user data directory is used, e.g. ~/.local/share/promnesia/
see https://github.com/ActiveState/appdirs#some-example-output for different OS
'''
OUTPUT_DIR = '/data/promnesia'

'''
Optional setting.
A directory to keep intemediate caches in order to speed up indexing.
If not specified, will use user cache directory
If set to None, cache is disabled
'''
CACHE_DIR = '/tmp/promnesia_cache/'


'''
Optional setting.
You can specify the URLs you don't want to be indexed.
You might want it because there are too many of them, or for security/privacy reasons.
'''
FILTERS = [
    'mail.google.com',
    'web.telegram.org/#/im',
    'vk.com/im',
    '192.168.0.',

    # you can use regexes too!
    'redditmedia.com.*.(jpg|png|gif)',
]

'''
Optional setting.
Can be useful to hack (e.g. rewrite/filter/etc) the visits before inserting in the database.
'''
def HOOK(v):
    # filter out 'boring' github visits without contexts
    if 'github.com' in v.norm_url:
        if v.context is None:
            return
    # otherwise keep intact
    yield v



'''
Uses [[https://github.com/karlicoss/HPI][HPI]] dogsheep module to import
Hacker News items.
'''

import textwrap

from promnesia.common import Visit, Loc, Results


def index() -> Results:
    from . import hpi
    from my.hackernews import dogsheep

    for item in dogsheep.items():
        if isinstance(item, Exception):
            yield item
            continue
        hn_url = item.permalink
        title = "hackernews"
        if item.title:
            title = item.title
        elif item.text_html:
            title = item.text_html
            title = textwrap.shorten(
                    title, width=79, placeholder="…",
                    break_long_words=True)
        # The locator is always the HN story. If the story is a link (as
        # opposed to a text post), we insert a visit such that the link
        # will point back to the corresponding HN story.
        loc = Loc.make(title=title, href=hn_url)
        urls = [hn_url]
        if item.url is not None:
            urls.append(item.url)
        for url in urls:
            yield Visit(
                    url=url,
                    dt=item.created,
                    locator=loc,
                    context=title,
            )



from datetime import datetime, timedelta
import os
from pathlib import Path
from subprocess import check_call, run, Popen
from textwrap import dedent
from typing import Set, Dict, Optional, Union, Sequence, Tuple, Mapping, List

import pytest

from common import tdir, under_ci, DATA, GIT_ROOT, promnesia_bin

from promnesia.common import _is_windows


def run_index(cfg: Path, *, update=False) -> None:
    from promnesia.__main__ import do_index
    do_index(cfg, overwrite_db=not update)


index = run_index # legacy name


def test_example_config(tmp_path: Path) -> None:
    if _is_windows:
        pytest.skip("doesn't work on Windows: example config references /usr/include paths")

    from promnesia.__main__ import read_example_config
    ex = read_example_config()
    cfg = tmp_path / 'test_config.py'
    cfg.write_text(ex)
    index(cfg)


from sqlalchemy import create_engine, MetaData, exists # type: ignore
from sqlalchemy import Column, Table # type: ignore
from cachew import NTBinder
from promnesia.common import DbVisit # TODO ugh. figure out pythonpath


# todo reuse promnesia.server??
def _get_stuff(outdir: Path):
    db_path = outdir / 'promnesia.sqlite'
    assert db_path.exists()

    engine = create_engine(f'sqlite:///{db_path}')

    binder = NTBinder.make(DbVisit)

    meta = MetaData(engine)
    table = Table('visits', meta, *binder.columns)

    return engine, binder, table

# TODO a bit shit... why did I make it dict at first??
Urls = Union[
           Mapping[str, Optional[str]],
    Sequence[Tuple[str, Optional[str]]],
]

def index_urls(urls: Urls, *, source_name: str='test'):
    uuu = list(urls.items()) if isinstance(urls, dict) else urls

    def idx(tdir: Path):
        cfg = tdir / 'test_config.py'
        cfg.write_text(f"""
OUTPUT_DIR = r'{tdir}'

from promnesia.common import Source, Visit, Loc
from datetime import datetime, timedelta
# todo reuse demo indexer?
indexer = Source(
    lambda: [Visit(
        url=url,
        dt=datetime.min + timedelta(days=5000) + timedelta(hours=i),
        locator=Loc.make('test'),
        context=ctx,
    ) for i, (url, ctx) in enumerate({uuu})],
    name=f'{source_name}',
)

SOURCES = [indexer]
""")
        index(cfg)
    return idx


def index_hypothesis(tdir: Path) -> None:
    hypexport_path  = DATA / 'hypexport'
    hypothesis_data = hypexport_path / 'testdata'

    cfg = tdir / 'test_config.py'
    # TODO ok, need to simplify this...
    cfg.write_text(f"""
OUTPUT_DIR = r'{tdir}'

from promnesia.common import Source

def hyp_extractor():
    import my.config
    class user_config:
        export_path = r'{str(hypothesis_data)}/netrights-dashboard-mockup/data/*.json'
    my.config.hypothesis = user_config

    # todo ideally would use virtualenv?
    import sys
    sys.path.insert(0, r"{str(hypexport_path / 'src')}")

    import promnesia.sources.hypothesis as hypothesis
    yield from hypothesis.index()

# in addition, test for lazy indexers. useful for importing packages
SOURCES = [
    Source(hyp_extractor, name='hyp'),
]
    """)
    index(cfg)


def index_some_demo_visits(
    tmp_path: Path,
    *,
    count: int,
    base_dt: datetime,
    delta: timedelta,
    update: bool,
) -> None:
    cfg = dedent(f'''
    OUTPUT_DIR = r'{tmp_path}'

    # todo would be nice if it was possible without importing anything at all
    def make_visits():
        from datetime import timedelta, datetime
        from promnesia.sources import demo

        # todo hmm, a bit messy, would be nice to do it in a more straighforward manner..
        yield from demo.index(
            count={count},
            base_dt=datetime.fromisoformat('{base_dt.isoformat()}'),
            delta=timedelta(seconds={delta.total_seconds()}),
        )

    from promnesia.common import Source
    SOURCES = [
        Source(make_visits, name='demo'),
    ]
    ''')
    cfg_file = tmp_path / 'test_config_extra.py'
    cfg_file.write_text(cfg)
    run_index(cfg_file, update=update)


def index_local_chrome(tdir: Path):
    # TODO mm, would be good to keep that for proper end2end
    # inp = Path('/L/data/promnesia/testdata/chrome-history/History') # TODO make it accessible to the repository
    # merged = tdir / 'chrome-merged.sqlite'
    # populate_db.merge_from('chrome', from_=inp, to=merged)

    merged = Path('/L/data/promnesia/testdata/chrome.sqlite')

    cfg = tdir / 'test_config.py'
    cfg.write_text(f"""
OUTPUT_DIR = r'{tdir}'

from promnesia.common import Indexer as I
# TODO FIXME -- fix back
from promnesia.sources.browser import chrome

chrome_extractor = I(chrome, '{merged}', src='chrome')

SOURCES = [chrome_extractor]
""")
    index(cfg)


def query_db_visits(tdir: Path) -> List[DbVisit]:
    # TODO copy pasting from server; need to unify
    engine, binder, table = _get_stuff(tdir)
    query = table.select()
    with engine.connect() as conn:
        return [binder.from_row(row) for row in conn.execute(query)]


def test_hypothesis(tdir: Path) -> None:
    index_hypothesis(tdir)
    visits = query_db_visits(tdir)
    assert len(visits) > 100

    [vis] = [x for x in visits if 'fundamental fact of evolution' in (x.context or '')]

    assert vis.norm_url == 'wired.com/2017/04/the-myth-of-a-superhuman-ai'
    assert vis.orig_url == 'https://www.wired.com/2017/04/the-myth-of-a-superhuman-ai/'
    assert vis.locator.href == 'https://hyp.is/_Z9ccmVZEeexBOO7mToqdg/www.wired.com/2017/04/the-myth-of-a-superhuman-ai/'
    assert 'misconception about evolution is fueling misconception about AI' in (vis.context or '') # contains notes as well


def test_comparison(tdir: Path) -> None:
    from promnesia.compare import compare_files

    idx = index_urls({
        'https://example.com': None,
        'https://en.wikipedia.org/wiki/Saturn_V': None,
        'https://plato.stanford.edu/entries/qualia': None,
    })
    idx(tdir)
    db     = tdir / 'promnesia.sqlite'
    old_db = tdir / 'promnesia-old.sqlite'
    import shutil
    shutil.move(str(db), str(old_db))

    idx2 = index_urls({
        'https://example.com': None,
        'https://www.reddit.com/r/explainlikeimfive/comments/1ev6e0/eli5entropy': None,
        'https://en.wikipedia.org/wiki/Saturn_V': None,
        'https://plato.stanford.edu/entries/qualia': None,
    })
    idx2(tdir)

    # should not crash, as there are more links in the new database
    assert len(list(compare_files(old_db, db))) == 0

    assert len(list(compare_files(db, old_db))) == 1


def test_index_many(tdir: Path) -> None:
    # NOTE [20200521] experimenting with promnesia.dump._CHUNK_BY
    # inserting 100K visits
    # value=1000: 9 seconds
    # value=10  : 9 seconds
    # value=1   : 18 seconds
    # ok, I guess it's acceptable considering the alternative is crashing (too many sql variables on some systems)
    # kinda makes sense -- I guess most overhead is coming from creating temporary lists etc?
    cfg = tdir / 'test_config.py'
    cfg.write_text(f"""
from datetime import datetime, timedelta
from promnesia.common import Source, Visit, Loc
# TODO def need to allow taking in index function without having to wrap in source?
def index():
    for i in range(100000):
        yield Visit(
            url='http://whatever/page' + str(i),
            dt=datetime.min + timedelta(days=5000) + timedelta(hours=i),
            locator=Loc.make('test'),
        )

SOURCES = [Source(index)]
OUTPUT_DIR = r'{tdir}'
    """)
    index(cfg)
    visits = query_db_visits(tdir)

    assert len(visits) == 100000


def test_indexing_error(tdir: Path) -> None:
    cfg = tdir / 'test_config.py'
    cfg.write_text(f'''
def bad_index():
    import bad_import
    yield from [] # dummy
from promnesia.common import Source
from promnesia.sources import demo

SOURCES = [
    bad_index,
    Source(demo, count=3),
]
OUTPUT_DIR = r'{tdir}'
''')
    with pytest.raises(SystemExit):
        index(cfg)
        # should exit(1)

    # yet save the database
    visits = query_db_visits(tdir)

    [e, _, _, _] = visits
    assert e.src == 'error'


def test_indexing_update(tmp_path: Path) -> None:
    from collections import Counter

    index_hypothesis(tmp_path)
    visits = query_db_visits(tmp_path)
    counter = Counter(v.src for v in visits)
    assert counter['hyp'] > 50, counter  # precondition

    dt = datetime.fromisoformat('2018-06-01T10:00:00.000000+01:00')
    index_some_demo_visits(tmp_path, count=1000, base_dt=dt, delta=timedelta(hours=1), update=True)
    visits = query_db_visits(tmp_path)
    counter = Counter(v.src for v in visits)
    assert counter['demo'] == 1000, counter
    assert counter['hyp'] > 50, counter  # should keep the original visits too!


@pytest.mark.parametrize('execution_number', range(1))  # adjust this parameter to increase 'coverage
def test_concurrent_indexing(tmp_path: Path, execution_number) -> None:
    cfg_slow = tmp_path / 'config_slow.py'
    cfg_fast = tmp_path / 'config_fast.py'
    cfg = dedent(f'''
    OUTPUT_DIR = r'{tmp_path}'
    from promnesia.common import Source
    from promnesia.sources import demo
    SOURCES = [Source(demo.index, count=COUNT)]
    ''')
    cfg_slow.write_text(cfg.replace('COUNT', '100000'))
    cfg_fast.write_text(cfg.replace('COUNT', '100'   ))
    # init it first, to create the database
    # TODO ideally this shouldn't be necessary but it's reasonable that people would already have the index
    # otherwise it would fail at db creation point.. which is kinda annoying to work around
    # todo in principle can work around same way as in cachew, by having a loop around PRAGMA WAL command?
    check_call(promnesia_bin('index', '--config', cfg_fast))

    # run in the background
    with Popen(promnesia_bin('index', '--config', cfg_slow)) as slow:
        while slow.poll() is None:
            # create a bunch of 'smaller' indexers running in parallel
            fasts = [Popen(promnesia_bin('index', '--config', cfg_fast)) for _ in range(10)]
            for fast in fasts:
                assert fast.wait() == 0, fast  # should succeed
        assert slow.poll() == 0, slow



#!/usr/bin/python3
__package__ = 'promnesia'  # ugh. hacky way to make hug work properly...

import argparse
import os
import sys
import json
from datetime import timedelta, datetime
from pathlib import Path
import logging
from functools import lru_cache
from typing import Collection, List, NamedTuple, Dict, Optional, Any, Tuple

from cachew import NTBinder

import pytz
from pytz import BaseTzInfo

import hug # type: ignore
import hug.types as T # type: ignore

from sqlalchemy import create_engine, MetaData, exists, literal, between, or_, and_, exc, select  # type: ignore
from sqlalchemy import Column, Table, func, types # type: ignore
from sqlalchemy.sql import text # type: ignore


from .common import PathWithMtime, DbVisit, Url, Loc, setup_logger, PathIsh, default_output_dir, get_system_tz
from .cannon import canonify


Json = Dict[str, Any]


# meh. need this since I don't have hooks in hug to initialize logging properly..
@lru_cache(1)
def get_logger() -> logging.Logger:
    # NOTE: uncomment to log sql queries
    # logging.basicConfig()
    # logging.getLogger('sqlalchemy.engine').setLevel(logging.DEBUG)

    # todo lazy log?
    logger = logging.getLogger('promnesia.server')
    setup_logger(logger, level=logging.DEBUG)

    # from hug.middleware import LogMiddleware
    # api = hug.API(__name__)
    # api.http.add_middleware(LogMiddleware(logger=logger))
    return logger


def get_version() -> str:
    from pkg_resources import get_distribution
    return get_distribution(__package__).version


class ServerConfig(NamedTuple):
    db: Path
    timezone: BaseTzInfo

    def as_str(self) -> str:
        return json.dumps({
            'timezone': self.timezone.zone,
            'db'      : str(self.db),
        })

    @classmethod
    def from_str(cls, cfgs: str) -> 'ServerConfig':
        d = json.loads(cfgs)
        return cls(
            db      =Path         (d['db']),
            timezone=pytz.timezone(d['timezone'])
        )


class EnvConfig:
    KEY = 'PROMNESIA_CONFIG'

    # apparently the only way to communicate with hug...
    @staticmethod
    @lru_cache(1)
    def get() -> ServerConfig:
        cfgs = os.environ.get(EnvConfig.KEY)
        assert cfgs is not None
        return ServerConfig.from_str(cfgs)

    @staticmethod
    def set(cfg: ServerConfig) -> None:
        os.environ[EnvConfig.KEY] = cfg.as_str()

# todo how to return exception in error?

def as_json(v: DbVisit) -> Json:
    # yep, this is NOT %Y-%m-%d as is seems to be the only format with timezone that Date.parse in JS accepts. Just forget it.
    dts = v.dt.strftime('%d %b %Y %H:%M:%S %z')
    loc = v.locator
    # TODO is locator always present??
    return {
        # TODO do not display year if it's current year??
        'dt': dts,
        # TODO the frontend had some bug with handling empty string as src. fix that later
        'src': v.src or 'unnamed',
        'context': v.context,
        'duration': v.duration,
        'locator': {
            'title': loc.title,
            'href' : loc.href,
        },
        'original_url'  : v.orig_url,
        'normalised_url': v.norm_url,
    }


def get_db_path(check: bool=True) -> Path:
    db = EnvConfig.get().db
    if check:
        assert db.exists(), db
    return db


@lru_cache(1)
# PathWithMtime aids lru_cache in reloading the sqlalchemy binder
def _get_stuff(db_path: PathWithMtime):
    get_logger().debug('Reloading DB: %s', db_path)
    # todo how to open read only?
    engine = create_engine(f'sqlite:///{db_path.path}') # , echo=True)

    binder = NTBinder.make(DbVisit)

    meta = MetaData(engine)
    table = Table('visits', meta, *binder.columns)

    from sqlalchemy import Index # type: ignore
    idx = Index('index_norm_url', table.c.norm_url)
    try:
        idx.create(bind=engine)
    except exc.OperationalError as e:
        if 'already exists' in str(e):
            # meh, but no idea how to check it properly...
            pass
        else:
            raise e

    # NOTE: apparently it's ok to open connection on every request? at least my comparisons didn't show anything
    return engine, binder, table


def get_stuff(db_path: Optional[Path]=None): # TODO better name
    # ok, it will always load from the same db file; but intermediate would be kinda an optional dump.
    if db_path is None:
        db_path = get_db_path()
    return _get_stuff(PathWithMtime.make(db_path))


def db_stats(*args, **kwargs) -> Json:
    engine, binder, table = get_stuff(*args, **kwargs)
    query = select([func.count()]).select_from(table)
    with engine.connect() as conn:
        total = list(conn.execute(query))[0][0]
    return {
        'total_visits': total,
    }


def search_common(url: str, where) -> Json:
    logger = get_logger()
    config = EnvConfig.get()

    logger.info('url: %s', url)
    original_url = url and url.strip()
    url = canonify(original_url)
    if not url:  # Don't eliminate a "#tag" query.
        url = original_url
    logger.info('normalised url: %s', url)

    visits0: List[Any] = []

    result = {
        'orginal_url'   : original_url,
        'normalised_url': url,
        'visits': visits0
    }

    engine, binder, table = get_stuff()

    query = table.select().where(where(table=table, url=url))
    logger.debug('query: %s', query)

    with engine.connect() as conn:
        try:
            # TODO make more defensive here
            visits: List[DbVisit] = [binder.from_row(row) for row in conn.execute(query)]
        except exc.OperationalError as e:
            if getattr(e, 'msg', None) == 'no such table: visits':
                logger.warn('you may have to run indexer first!')
                #result['visits'] = [{an error with a msg}] # TODO
                #return result
            raise

    logger.debug('got %d visits from db', len(visits))

    vlist: List[DbVisit] = []
    for vis in visits:
        dt = vis.dt
        if dt.tzinfo is None: # FIXME need this for /visits endpoint as well?
            tz = config.timezone
            dt = tz.localize(dt)
            vis = vis._replace(dt=dt)
        vlist.append(vis)

    logger.debug('responding with %d visits', len(vlist))
    # TODO respond with normalised result, then frontent could choose how to present children/siblings/whatever?
    result['visits'] = list(map(as_json, vlist))
    return result


@hug.local()
@hug.get ('/status')
# NOTE: not sure why I used post in the first place... but it was used in the extension so need to keep
@hug.post('/status')
def status():
    '''
    Ideally, status will always respond, regardless the internal state of the backend?
    '''
    # TODO hug stats?
    logger = get_logger()

    db = get_db_path(check=False)
    try:
        assert db.exists(), db
        db_path = str(db)
    except Exception as e:
        logger.exception(e)
        db_path = f'ERROR: db not found/unreadable (expected path {db}). You probably forgot to run indexer first. See https://github.com/karlicoss/promnesia/blob/master/doc/TROUBLESHOOTING.org'

    stats: Json
    try:
        stats = db_stats(db)
    except Exception as e:
        logger.exception(e)
        stats = {'ERROR': str(e)}

    version: Optional[str]
    try:
        version = get_version()
    except Exception as e:
        version = None

    return {
        'version': version,
        'db'     : db_path,
        'stats'  : stats,
    }


# TODO ugh. hug doesn't like return Json type??
# fails with "TypeError: Type Dict cannot be instantiated; use dict() instead"
@hug.local()
@hug.post('/visits')
def visits(
        url: T.text,
):
    get_logger().info('/visited %s', url)
    return search_common(
        url=url,
        # odd, doesn't work just with: x or (y and z)
        where=lambda table, url: or_(
            table.c.norm_url == url,  # exact match
            and_(table.c.context != None, table.c.norm_url.startswith(url, autoescape=True)) # + child visits, but only 'interesting' ones
        ),
    )


@hug.local()
@hug.post('/search')
def search(
        url: T.text
):
    get_logger().info('/search %s', url)
    return search_common(
        url=url,
        where=lambda table, url: or_(
            # todo hmm. think about it, not sure if I need proper indexer for fuzzy search etc?
            table.c.norm_url     .contains(url, autoescape=True),
            table.c.orig_url     .contains(url, autoescape=True),
            table.c.context      .contains(url, autoescape=True),
            table.c.locator_title.contains(url, autoescape=True),
        ),
    )


@hug.local()
@hug.post('/search_around')
def search_around(
        timestamp: T.number,
):
    get_logger().info('/search_around %s', timestamp)
    utc_timestamp = timestamp # old 'timestamp' name is legacy

    # TODO meh. use count/pagination instead?
    delta_back  = timedelta(hours=3  ).total_seconds()
    delta_front = timedelta(minutes=2).total_seconds()
    # TODO not sure about delta_front.. but it also serves as quick hack to accommodate for all the truncations etc

    return search_common(
        url='http://dummy.org', # NOTE: not used in the where query (below).. perhaps need to get rid of this
        where=lambda table, url: between(
            func.strftime(
                '%s', # NOTE: it's tz aware, e.g. would distinguish +05:00 vs -03:00
                # this is a bit fragile, relies on cachew internal timestamp format, e.g.
                # 2020-11-10T06:13:03.196376+00:00 Europe/London
                func.substr(
                    table.c.dt,
                    1, # substr is 1-indexed
                    # instr finds the first match, but if not found it defaults to 0.. which we hack by concatting with ' '
                    func.instr(func.cast(table.c.dt, types.Unicode).op('||')(' '), ' ') - 1,
                    # for fucks sake.. seems that cast is necessary otherwise it tries to treat ' ' as datetime???
                )
            ) - literal(utc_timestamp),
            literal(-delta_back),
            literal(delta_front),
        ),
    )

# before 0.11.14 (including), extension didn't share the version
# so if it's not shared, assume that version
_NO_VERSION = (0, 11, 14)
_LATEST = (9999, 9999, 9999)

def as_version(version: str) -> Tuple[int, int, int]:
    if version == '':
        return _NO_VERSION
    try:
        [v1, v2, v3] = map(int, version.split('.'))
        return (v1, v2, v3)
    except Exception as e:
        logger = get_logger()
        logger.error('error while parsing version %s', version)
        logger.exception(e)
        return _LATEST


# todo ugh. hug chokes over annotation types? List[Dict[str, Any]]
@hug.local()
@hug.post('/visited')
def visited(
        urls,
        client_version: str='',
):
    logger = get_logger()
    logger.info('/visited %s %s', urls, client_version)

    version = as_version(client_version)

    nurls = [canonify(u) for u in urls]
    snurls = list(sorted(set(nurls)))

    if len(snurls) == 0:
        return []

    engine, binder, table = get_stuff()

    # sqlalchemy doesn't seem to support SELECT FROM (VALUES (...)) in its api
    # also doesn't support array binding...
    # https://stackoverflow.com/questions/13190392/how-can-i-bind-a-list-to-a-parameter-in-a-custom-query-in-sqlalchemy
    bstring = ','.join(f'(:b{i})'   for i, _ in enumerate(snurls))
    bdict = {            f'b{i}': v for i, v in enumerate(snurls)}
    # TODO hopefully, visits.* thing only returns one visit??
    query = text(f"""
WITH cte(queried) AS (SELECT * FROM (values {bstring}))
SELECT queried, visits.*
    FROM cte JOIN visits
    ON queried = visits.norm_url
/*  order stuff without contexts last
    this actually doesn't make sense, locially it should be ASC??
    but somehow DESC is the one that actually works..
*/
    ORDER BY visits.context IS NULL DESC
    """).bindparams(**bdict).columns(
        Column('match', types.Unicode),
        *table.columns,
    )
    # TODO might be very beneficial for performance to have an intermediate table
    # SELECT visits.* FROM visits GROUP BY visits.norm_url ORDER BY visits.context IS NULL DESC
    # + unique index in norm_url
    # brings down large queries to 50ms...
    with engine.connect() as conn:
        res = list(conn.execute(query))
        present = {row[0]: binder.from_row(row[1:]) for row in res}
    results = []
    for nu in nurls:
        r = present.get(nu, None)
        results.append(None if r is None else as_json(r))

    if version <= (0, 11, 14):
        # older extension versions expected boolean result here
        results = [r is not None for r in results] # type: ignore[misc]

    return results


def _run(*, port: str, quiet: bool, config: ServerConfig) -> None:
    logger = get_logger()

    logger.info('Running hug with %s', config)

    EnvConfig.set(config)
    import hug.development_runner # type: ignore
    hug.development_runner.hug(
        file=__file__,
        port=int(port),
    )


def run(args: argparse.Namespace) -> None:
    _run(
        port=args.port,
        quiet=args.quiet,
        config=ServerConfig(
            db=args.db,
            timezone=args.timezone,
        )
    )


def default_db_path() -> Path:
    return default_output_dir() / 'promnesia.sqlite'


def setup_parser(p: argparse.ArgumentParser) -> None:
    p.add_argument(
        '--port',
        type=str,
        default='13131',
        help='Port for communicating with extension',
    )

    p.add_argument(
        '--quiet',
        action='store_true',
        help='Pass to log less',
    )
    # TODO need to keep consistent with the backend...
    # todo use output_dir instead?
    p.add_argument(
        '--db',
        type=Path,
        default=default_db_path(),
        help='Path to the links database (optional, uses user data dir by default)',
    )

    p.add_argument(
        '--timezone',
        type=pytz.timezone,
        default=get_system_tz(),
        help='Fallback timezone, defaults to the system timezone if not specified',
    )



from pathlib import Path
import os
from typing import List, Optional, Union, NamedTuple, Iterable, Callable
import importlib
import importlib.util
import warnings

from .common import PathIsh, get_tmpdir, appdirs, default_output_dir, default_cache_dir, user_config_file
from .common import Res, Source, DbVisit


HookT = Callable[[Res[DbVisit]], Iterable[Res[DbVisit]]]

class Config(NamedTuple):
    # TODO remove default from sources once migrated
    SOURCES: List = []

    # if not specified, uses user data dir
    OUTPUT_DIR: Optional[PathIsh] = None

    CACHE_DIR: Optional[PathIsh] = ''
    FILTERS: List[str] = []

    HOOK: Optional[HookT] = None

    #
    # NOTE: INDEXERS is deprecated, use SOURCES instead
    INDEXERS: List = []
    #MIME_HANDLER: Optional[str] = None # TODO
    @property
    def sources(self) -> Iterable[Res[Source]]:
        idx = self.INDEXERS

        if len(self.INDEXERS) > 0:
            warnings.warn("'INDEXERS' is deprecated. Please use 'SOURCES'!", DeprecationWarning)

        raw = self.SOURCES + self.INDEXERS

        if len(raw) == 0:
            raise RuntimeError("Please specify SOURCES in the config! See https://github.com/karlicoss/promnesia#setup for more information")


        for r in raw:
            if isinstance(r, str):
                # must be a raw module name?
                try:
                    r = importlib.import_module(r)
                except ModuleNotFoundError as e:
                    yield e
                    continue

            if isinstance(r, Source):
                yield r
            else:
                # kind of last resort measure..
                yield Source(r)

    @property
    def cache_dir(self) -> Optional[Path]:
        cd = self.CACHE_DIR
        cpath: Optional[Path]
        if cd is None:
            cpath = None # means 'disabled' in cachew
        elif cd == '': # meh.. but need to make it None friendly..
            cpath = default_cache_dir()
        else:
            cpath = Path(cd)
        if cpath is not None:
            cpath.mkdir(exist_ok=True, parents=True)
        return cpath

    # TODO also tmp dir -- perhaps should be in cache or at least possible to specify in config? not sure if useful
    @property
    def output_dir(self) -> Path:
        odir = self.OUTPUT_DIR
        opath = default_output_dir() if odir is None else Path(odir)
        opath.mkdir(exist_ok=True, parents=True)
        return opath

    @property
    def db(self) -> Path:
        return self.output_dir / 'promnesia.sqlite'

    @property
    def hook(self) -> Optional[HookT]:
        return self.HOOK

instance: Optional[Config] = None


def has() -> bool:
    return instance is not None

def get() -> Config:
    assert instance is not None, "Expected config to be set, but it's not"
    return instance


def load_from(config_file: Path) -> None:
    global instance
    instance = import_config(config_file)


def reset() -> None:
    global instance
    assert instance is not None
    instance = None


def import_config(config_file: PathIsh) -> Config:
    p = Path(config_file)

    # todo just exec??
    name = p.stem
    spec = importlib.util.spec_from_file_location(name, p); assert spec is not None
    mod = importlib.util.module_from_spec(spec); assert mod is not None
    loader = spec.loader; assert loader is not None
    loader.exec_module(mod) # type: ignore[attr-defined]

    d = {}
    for f in Config._fields:
        if hasattr(mod, f):
            d[f] = getattr(mod, f)
    return Config(**d)


# TODO: ugh. this causes warnings to be repeated multiple times... need to reuse the pool or something..
def use_cores() -> Optional[int]:
    '''
    Somewhat experimental.
    For now only used in sources.auto, perhaps later will be shared among the other indexers.
    '''
    # most likely needs to be some sort of pipeline thing?
    cs = os.environ.get('PROMNESIA_CORES', None)
    if cs is None:
        return None
    try:
        return int(cs)
    except ValueError: # any other value means 'use all
        return 0


def extra_fd_args() -> List[str]:
    '''
    Not sure where it belongs yet... so via env variable for now
    Can be used to pass --ignore-file parameter
    '''
    v = os.environ.get('PROMNESIA_FD_EXTRA_ARGS', '')
    extra = v.split() # eh, hopefully splitting that way is ok...
    return extra



#!/usr/bin/env python3
from contextlib import contextmanager
import json
from pathlib import Path
from datetime import datetime
from tempfile import TemporaryDirectory
import os
from subprocess import check_call, check_output
from time import sleep
from typing import NamedTuple, Optional, Iterator

import pytest # type: ignore

if __name__ == '__main__':
    # TODO ugh need to figure out PATH
    # python3 -m pytest -s tests/server_test.py::test_query
    pytest.main(['-s', __file__])


from selenium import webdriver # type: ignore
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait as Wait
from selenium.webdriver.support import expected_conditions as EC


from common import under_ci, uses_x, has_x
from integration_test import index_hypothesis, index_local_chrome, index_urls
from server_test import wserver
from browser_helper import open_extension_page, get_cmd_hotkey

from promnesia.logging import LazyLogger


logger = LazyLogger('promnesia-tests')


class Browser(NamedTuple):
    dist: str
    headless: bool

    @property
    def name(self) -> str:
        return self.dist.split('-')[0] # TODO meh

    def skip_ci_x(self):
        if under_ci() and not self.headless:
            pytest.skip("Only can't use headless browser on CI")


FF  = Browser('firefox', headless=False)
CH  = Browser('chrome' , headless=False)
FFH = Browser('firefox', headless=True)
# NOTE: sadly headless chrome doesn't support extensions..
# https://stackoverflow.com/a/45372648/706389
# there is some workaround, but it's somewhat tricky...
# https://stackoverflow.com/a/46475980/706389


# TODO ugh, I guess it's not that easy to make it work because of isAndroid checks...
# I guess easy way to test if you really want is to temporary force isAndroid to return true in extension...
FM  = Browser('firefox-mobile', headless=False)


def browser_(driver):
    name = driver.name
    # TODO figure out headless??
    if name == 'firefox':
        return FF
    elif name == 'chrome':
        return CH
    else:
        raise AssertionError(driver)


def get_addon_path(kind: str) -> Path:
    # TODO compile first?
    addon_path = (Path(__file__).parent.parent / 'extension' / 'dist' / kind).absolute()
    assert addon_path.exists()
    assert (addon_path / 'manifest.json').exists()
    return addon_path


def get_hotkey(driver, cmd: str) -> str:
    cmd_map = None
    if driver.name == 'firefox':
        # TODO remove hardcoding somehow...
        # perhaps should be extracted somewhere..
        cmd_map = {
            Command.MARK_VISITED     : 'Ctrl+Alt+v',
            '_execute_browser_action': 'Ctrl+Alt+e',
            'search'                 : 'Ctrl+Alt+h',
        }
    return get_cmd_hotkey(driver, cmd=cmd, cmd_map=cmd_map)


def _get_webdriver(tdir: Path, browser: Browser, extension=True):
    addon = get_addon_path(kind=browser.dist)
    if browser.name == 'firefox':
        options = webdriver.FirefoxOptions()
        options.set_preference('profile', str(tdir))
        options.headless = browser.headless
        # use firefox from here to test https://www.mozilla.org/en-GB/firefox/developer/
        driver = webdriver.Firefox(options=options)
        # todo pass firefox-dev binary?
        if extension:
            driver.install_addon(str(addon), temporary=True)
    elif browser.name == 'chrome':
        # TODO ugh. very hacky...
        assert extension, "TODO add support for extension arg"
        ex = tdir / 'extension.zip'
        files = [x.name for x in addon.iterdir()]
        check_call(['apack', '-q', str(ex), *files], cwd=addon)
        # looks like chrome uses temporary dir for data anyway
        options = webdriver.ChromeOptions()
        options.headless = browser.headless
        options.add_extension(ex)
        driver = webdriver.Chrome(options=options)
    else:
        raise RuntimeError(f'Unexpected browser {browser}')
    return driver


# TODO copy paste from grasp
@contextmanager
def get_webdriver(browser: Browser, extension=True):
    with TemporaryDirectory() as td:
        tdir = Path(td)
        driver = _get_webdriver(tdir, browser=browser, extension=extension)
        try:
            yield driver
        finally:
            driver.quit()


def set_host(*, driver, host: Optional[str], port: Optional[str]) -> None:
    # todo rename to 'backend_id'?
    if host is None:
        ep = driver.find_element(By.ID, 'host_id')
        ep.clear()
        return
    assert port is not None
    ep = driver.find_element(By.ID, 'host_id')
    ep.clear()
    ep.send_keys(f'{host}:{port}')


def save_settings(driver):
    se = driver.find_element(By.ID, 'save_id')
    se.click()

    driver.switch_to.alert.accept()


LOCALHOST = 'http://localhost'


def configure_extension(
        driver,
        *,
        host: Optional[str]=LOCALHOST,
        port: Optional[str]=None,
        show_dots: bool=True,
        highlights: Optional[bool]=None,
        blacklist=None,
        excludelists=None,
        notify_contexts: Optional[bool]=None,
        position: Optional[str]=None,
        verbose_errors: bool=True,
) -> None:
    def set_checkbox(cid: str, value: bool):
        cb = driver.find_element(By.ID, cid)
        selected = cb.is_selected()
        if selected != value:
            cb.click()


    # TODO log properly
    print(f"Setting: port {port}, show_dots {show_dots}")

    open_extension_page(driver, page='options_page.html')
    sleep(1) # err, wtf? otherwise not always interacts with the elements correctly

    set_host(driver=driver, host=host, port=port)

    # dots = driver.find_element(By.ID, 'dots_id')
    # if dots.is_selected() != show_dots:
    #     dots.click()
    # assert dots.is_selected() == show_dots

    # TODO not sure, should be False for demos?
    set_checkbox('verbose_errors_id', verbose_errors)

    if highlights is not None:
        set_checkbox('highlight_id', highlights)

    if notify_contexts is not None:
        set_checkbox('contexts_popup_id', notify_contexts)

    if position is not None:
        set_position(driver, position)

    if blacklist is not None:
        bl = driver.find_element(By.ID, 'global_excludelist_id') # .find_element_by_tag_name('textarea')
        bl.click()
        # ugh, that's hacky. presumably due to using Codemirror?
        bla = driver.switch_to.active_element
        # ugh. doesn't work, .text always returns 0...
        # if len(bla.text) > 0:
        # for some reason bla.clear() isn't working...
        # results in ElementNotInteractableException: Element <textarea> could not be scrolled into view
        bla.send_keys(Keys.CONTROL + 'a')
        bla.send_keys(Keys.BACKSPACE)
        bla.send_keys('\n'.join(blacklist))

    if excludelists is not None:
        excludelists_json = json.dumps(excludelists)

        bl = driver.find_element(By.ID, 'global_excludelists_ext_id') # .find_element_by_tag_name('textarea')
        bl.click()
        # ugh, that's hacky. presumably due to using Codemirror?
        bla = driver.switch_to.active_element
        # ugh. doesn't work, .text always returns 0...
        # if len(bla.text) > 0:
        # for some reason bla.clear() isn't working...
        # results in ElementNotInteractableException: Element <textarea> could not be scrolled into view
        bla.send_keys(Keys.CONTROL + 'a')
        bla.send_keys(Keys.BACKSPACE)
        bla.send_keys(excludelists_json)

    save_settings(driver)


# legacy name
configure = configure_extension


def get_window_id(driver):
    if driver.name == 'firefox':
        pid = str(driver.capabilities['moz:processID'])
    else:
        # ugh nothing in capabilities...
        pid = check_output(['pgrep', '-f', 'chrome.*enable-automation']).decode('utf8').strip()
    # https://askubuntu.com/a/385037/427470
    return get_wid_by_pid(pid)


def get_wid_by_pid(pid: str):
    wids = check_output(['xdotool', 'search', '--pid', pid]).decode('utf8').splitlines()
    wids = [w.strip() for w in wids if len(w.strip()) > 0]

    def has_wm_desktop(wid: str) -> bool:
        # TODO no idea why is that important. found out experimentally
        out = check_output(['xprop', '-id', wid, '_NET_WM_DESKTOP']).decode('utf8')
        return 'not found' not in out

    [wid] = filter(has_wm_desktop, wids)
    return wid


def focus_browser_window(driver):
    wid = get_window_id(driver)
    check_call(['xdotool', 'windowactivate', wid])


def trigger_callback(driver, callback):
    focus_browser_window(driver)
    sleep(0.5)
    callback()


def send_key(key):
    if isinstance(key, str):
        key = key.split('+')

    print(f"sending hotkey! {key}")
    import pyautogui # type: ignore
    pyautogui.hotkey(*key)


def is_headless(driver) -> bool:
    if driver.name == 'firefox':
        return driver.capabilities.get('moz:headless', False)
    else:
        return False


# TODO move to common or something
def trigger_hotkey(driver, hotkey: str) -> None:
    headless = is_headless(driver)
    try:
        trigger_callback(driver, lambda: send_key(hotkey))
    except Exception as e:
        if headless:
            raise RuntimeError("Likely failed because the browser is headless") from e
        else:
            raise e


def trigger_command(driver, cmd: str) -> None:
    if is_headless(driver):
        assert cmd == Command.ACTIVATE, cmd  # will support the rest later..
        # see background-injector.js
        driver.execute_script("""
        var event = document.createEvent('HTMLEvents');
        event.initEvent('selenium-bridge-activate', true, true);
        document.dispatchEvent(event);
        """)
    else:
        trigger_hotkey(driver, get_hotkey(driver, cmd))


class TestHelper(NamedTuple):
    driver: webdriver.Remote

    def open_page(self, page: str) -> None:
        open_extension_page(self.driver, page)

    def move_to(self, element):
        from selenium.webdriver.common.action_chains import ActionChains # type: ignore
        ActionChains(self.driver).move_to_element(element).perform()

    def switch_to_sidebar(self, wait=False):
        self.driver.switch_to.default_content()

        frame_id = 'promnesia-frame'
        frame = None
        if wait:
            Wait(self.driver, timeout=5).until(
                EC.presence_of_element_located((By.ID, frame_id))
            )
        self.driver.switch_to.frame(frame_id)

    @contextmanager
    def sidebar(self, wait=False):
        try:
            self.switch_to_sidebar(wait=wait)
            if wait: # meh
                yield self.driver.find_element_by_id('promnesia-sidebar')
            else:
                yield None
        finally:
            self.driver.switch_to.default_content()

    def command(self, cmd):
        trigger_command(self.driver, cmd)

    def activate(self) -> None:
        self.command(Command.ACTIVATE)

    def mark_visited(self) -> None:
        self.command(Command.MARK_VISITED)

    def wid(self) -> str:
        return get_window_id(self.driver)

    def screenshot(self, path):
        # ugh, webdriver's save_screenshot doesn't behave well with frames
        check_call(['import', '-window', self.wid(), path])


def confirm(what: str) -> None:
    is_headless = 'headless' in os.environ.get('PYTEST_CURRENT_TEST', '')
    if is_headless:
        # ugh.hacky
        Headless().confirm(what)
        return

    import click # type: ignore
    click.confirm(click.style(what, blink=True, fg='yellow'), abort=True)


class Manual:
    def confirm(self, what: str) -> None:
        raise NotImplementedError

class Interactive(Manual):
    def confirm(self, what: str) -> None:
        confirm(what)

class Headless(Manual):
    def confirm(self, what: str) -> None:
        logger.warning('"%s": headless mode, responding "yes"', what)


'''
Helper for tests that are not yet fully automated and require a human to check...
- if running with the GUI, will be interactive
- if running in headless mode, will automatically assume 'yes'.
  of course it's not very robust, but at least we're testing some codepaths then
'''
manual = Interactive() if has_x() else Headless()


@contextmanager
def _test_helper(tmp_path, indexer, test_url: Optional[str], browser: Browser=FFH, **kwargs):
    tdir = Path(tmp_path)

    indexer(tdir)
    with wserver(db=tdir / 'promnesia.sqlite') as srv, get_webdriver(browser=browser) as driver:
        port = srv.port
        configure(driver, port=port, **kwargs)
        sleep(0.5)

        if test_url is not None:
            driver.get(test_url)
            sleep(3) # todo use some condition...
        else:
            driver.get('about:blank')

        yield TestHelper(driver=driver)

class Command:
    MARK_VISITED = 'mark_visited'
    ACTIVATE  = '_execute_browser_action'
    SEARCH    = 'search'
# TODO assert this against manifest?


WITH_BROWSER_TESTS = 'WITH_BROWSER_TESTS'

with_browser_tests = pytest.mark.skipif(
    WITH_BROWSER_TESTS not in os.environ,
    reason=f'set env var {WITH_BROWSER_TESTS}=true if you want to run this test',
)


def browsers(*br: Browser):
    if len(br) == 0:
        br = (FF, FFH, CH)
    if not has_x():
        br = tuple(b for b in br if b.headless)

    from functools import wraps
    def dec(f):
        if len(br) == 0:
            dec_ = pytest.mark.skip('Filtered out all browsers (because of no GUI/non-interactive mode)')
        else:
            dec_ = pytest.mark.parametrize('browser', br, ids=lambda b: b.dist.replace('-', '_') + ('_headless' if b.headless else ''))
        @with_browser_tests
        @dec_
        @wraps(f)
        def ff(*args, **kwargs):
            return f(*args, **kwargs)
        return ff
    return dec


PYTHON_DOC_URL = 'file:///usr/share/doc/python3/html/index.html'


@browsers()
def test_installs(tmp_path, browser):
    with get_webdriver(browser=browser):
        # just shouldn't crash
        pass


@browsers()
def test_settings(tmp_path, browser):
    # TODO fixture for driver?
    with get_webdriver(browser=browser) as driver:
        configure_extension(driver, port='12345', show_dots=False)
        driver.get('about:blank')
        open_extension_page(driver, page='options_page.html')
        hh = driver.find_element(By.ID, 'host_id')
        assert hh.get_attribute('value') == 'http://localhost:12345'


@browsers()
def test_backend_status(tmp_path, browser):
    with get_webdriver(browser=browser) as driver:
        helper = TestHelper(driver)
        helper.open_page('options_page.html')
        sleep(1) # ugh. for some reason pause here seems necessary..

        set_host(driver=driver, host='https://nosuchhost.com', port='1234')
        driver.find_element(By.ID, 'backend_status_id').click()
        sleep(1 + 0.5) # needs enough time for timeout to trigger...

        alert = driver.switch_to.alert
        assert 'ERROR' in alert.text

        sleep(0.5) #  don't remember if needed?
        driver.switch_to.alert.accept()
        # TODO implement positive check??

def set_position(driver, settings: str):
    browser = browser_(driver)

    # TODO figure out browser from driver??
    field = driver.find_element_by_xpath('//*[@id="position_css_id"]')
    if browser.name == 'chrome':
        # ugh... for some reason wouldn't send the keys...
        field.click()
        import pyautogui # type: ignore
        # it usually ends up in the middle of the area...
        pyautogui.press(['backspace'] * 500 + ['delete'] * 500)
        pyautogui.typewrite(settings, interval=0.05)
    else:
        area = field.find_element_by_xpath('.//textarea')
        area.send_keys([Keys.DELETE] * 1000)
        area.send_keys(settings)


@browsers()
def test_sidebar_bottom(browser):
    with get_webdriver(browser=browser) as driver:
        open_extension_page(driver, page='options_page.html')
        sleep(1) # ugh. for some reason pause here seems necessary..

        # for some reason area.clear() caused
        # selenium.common.exceptions.ElementNotInteractableException: Message: Element <textarea> could not be scrolled into view

        settings = """
#promnesia-sidebar {
    --bottom: 1;
    --size: 20%;
}"""

        set_position(driver, settings)

        save_settings(driver)

        driver.get(PYTHON_DOC_URL)

        trigger_command(driver, Command.ACTIVATE)
        confirm("You should see sidebar below")


@browsers()
def test_blacklist_custom(tmp_path: Path, browser: Browser) -> None:
    with get_webdriver(browser=browser) as driver:
        helper = TestHelper(driver)
        configure_extension(driver, port='12345', blacklist=('stackoverflow.com',))
        driver.get('https://stackoverflow.com/questions/27215462')

        helper.activate()

        manual.confirm('page should be blacklisted (black icon), your should see an error notification')

        # make sure we can't open sidebar for blacklisted page
        # meh, but that'll do for now
        from selenium.common.exceptions import NoSuchFrameException
        with pytest.raises(NoSuchFrameException):
            with helper.sidebar():
                pass

        # reset blacklist
        # also running without backend here, so need to set host to none as well
        configure_extension(driver, host=None, blacklist=())
        driver.back() # TODO maybe configure_extension should go back instead...
        driver.refresh()

        with helper.sidebar(wait=True):
            # at least shouln't fail
            pass

        manual.confirm('sidebar should be visible')


@browsers()
def test_blacklist_builtin(tmp_path: Path, browser: Browser) -> None:
    with get_webdriver(browser=browser) as driver:
        helper = TestHelper(driver)
        configure_extension(driver, port='12345')
        driver.get('https://www.hsbc.co.uk/mortgages/')

        helper.activate()

        manual.confirm('page should be blacklisted (black icon), your should see an error notification')

        # make sure we can't open sidebar for blacklisted page
        # meh, but that'll do for now
        from selenium.common.exceptions import NoSuchFrameException
        with pytest.raises(NoSuchFrameException):
            with helper.sidebar():
                pass

        # reset blacklist
        # also running without backend here, so need to set host to none as well
        configure_extension(driver, host=None, excludelists=())
        driver.back() # TODO maybe configure_extension should go back instead...
        driver.refresh()

        with helper.sidebar(wait=True):
            # at least shouln't fail
            pass

        manual.confirm('sidebar should be visible')


@browsers(FF, CH)
def test_add_to_blacklist(tmp_path, browser):
    with get_webdriver(browser=browser) as driver:
        configure_extension(driver, port='12345')
        driver.get('https://example.com')
        chain = webdriver.ActionChains(driver)
        chain.move_to_element(driver.find_element(By.TAG_NAME, 'h1')).context_click().perform()

        # looks like selenium can't interact with browser context menu...
        import pyautogui # type: ignore

        if driver.name == 'chrome':
            offset = 2 # Inspect, View page source
        else:
            offset = 0
        pyautogui.typewrite(['up'] + ['up'] * offset + ['enter'] + ['enter'], interval=0.5)

        confirm('shows prompt with alert to enter pattern to block?')
        driver.switch_to.alert.accept()
        # ugh, seems necessary to guard with sleep; otherwise racey
        sleep(0.5)

        driver.get(driver.current_url)
        confirm('page should be blacklisted (black icon)')



# todo might be nice to run soft asserts for this test?
@browsers()
def test_visits(tmp_path: Path, browser: Browser) -> None:
    test_url = "http://www.e-flux.com/journal/53/59883/the-black-stack/"
    # test_url = "file:///usr/share/doc/python3/html/library/contextlib.html" # TODO ??
    with _test_helper(tmp_path, index_hypothesis, test_url, browser=browser) as helper:
        driver = helper.driver
        with helper.sidebar():
            # hmm not sure how come it returns anything at all.. but whatever
            srcs = driver.find_elements_by_class_name('src')
            for s in srcs:
                assert not s.is_displayed(), s
            assert len(srcs) >= 8, srcs
            # todo ugh, need to filter out filters, how to only query the ones in the sidebar?

        helper.activate()

        with helper.sidebar(wait=True) as sidebar:
            assert sidebar is not None

            confirm('you should see hypothesis contexts')

            link = driver.find_element_by_partial_link_text('how_algorithms_shape_our_world')
            assert link.is_displayed(), link

            contexts = helper.driver.find_elements_by_class_name('context')
            for c in contexts:
                assert c.is_displayed(), c
            assert len(contexts) == 8


        helper.activate()
        # todo wait till it disappears properly?
        sleep(0.5)

        with helper.sidebar():
            # hmm not sure how come it returns anything at all.. but whatever
            srcs = driver.find_elements_by_class_name('src')
            for s in srcs:
                assert not s.is_displayed(), s


@browsers()
def test_search_around(tmp_path, browser) -> None:
    # TODO it actually lacks a proper end-to-end test within browser... although I do have something in automatic demos?
    test_url = "about:blank"
    with _test_helper(tmp_path, index_hypothesis, test_url, browser=browser) as h:
        # TODO hmm. dunno if we want to highlight only result with the same timestamp, or the results that are 'near'??
        ts = int(datetime.strptime("2017-05-22T10:59:12.082375+00:00", '%Y-%m-%dT%H:%M:%S.%f%z').timestamp())
        h.open_page(f'search.html?utc_timestamp_s={ts}')
        manual.confirm('you should see search results, "anthrocidal" should be highlighted red')
        # FIXME test clicking search around in actual search page.. it didn't work, seemingly because of initBackground() handling??


# TODO skip if not my hostname
@uses_x
@browsers(FF, CH)
def test_chrome_visits(tmp_path, browser):
    test_url = "https://en.wikipedia.org/wiki/Amplituhedron"
    test_url = "https://en.wikipedia.org/wiki/Symplectic_vector_space"
    with _test_helper(tmp_path, index_local_chrome, test_url, browser=browser) as helper:
        trigger_command(helper.driver, Command.ACTIVATE)
        confirm("You shoud see chrome visits now; with time spent")


@browsers(FF, CH)
def test_show_visited_marks(tmp_path, browser):
    visited = {
        'https://en.wikipedia.org/wiki/Special_linear_group': None,
        'http://en.wikipedia.org/wiki/Unitary_group'        : None,
        'en.wikipedia.org/wiki/Transpose'                   : None,
    }
    test_url = "https://en.wikipedia.org/wiki/Symplectic_group"
    with _test_helper(tmp_path, index_urls(visited), test_url, show_dots=True, browser=browser) as helper:
        trigger_command(helper.driver, Command.MARK_VISITED)
        confirm("You should see visited marks near special linear group, Unitary group, Transpose")


@browsers(FF, CH)
@pytest.mark.parametrize(
    'url',
    [
        "https://en.wikipedia.org/wiki/Symplectic_group",

        # regression test for https://github.com/karlicoss/promnesia/issues/295
        # note: seemed to reproduce on chrome more consistently for some reason
        "https://www.udemy.com/course/javascript-bible/",
    ],
    ids=[
        'wiki',
        'udemy'
    ],
)
def test_sidebar(tmp_path, browser, url: str) -> None:
    visited = {
        url : 'whatever',
    }
    indexer = index_urls(visited, source_name='ælso test unicode 💩')
    with _test_helper(tmp_path, indexer, url, show_dots=True, browser=browser) as helper:
        trigger_command(helper.driver, Command.ACTIVATE)
        confirm("You should see green icon, also one visit in sidebar. Make sure the unicode is displayed correctly.")


@browsers(FF, CH)
def test_search(tmp_path, browser):
    test_url = "https://en.wikipedia.org/wiki/Symplectic_vector_space"
    with _test_helper(tmp_path, index_hypothesis, test_url, browser=browser) as helper:
        trigger_command(helper.driver, Command.SEARCH)
        # TODO actually search something?
        # TODO use current domain as default? or 'parent' url?
        confirm("You shoud see search prompt now, with focus on search field")


@browsers()
def test_new_background_tab(tmp_path, browser) -> None:
    start_url = "http://www.e-flux.com/journal/53/59883/the-black-stack/"
    # bg_url_text = "El Proceso (The Process)"
    # TODO generate some fake data instead?
    with _test_helper(
            tmp_path, index_hypothesis, start_url, browser=browser,
            notify_contexts=True,
    ) as helper:
        manual.confirm('you should see notification about contexts')
        page_logo = helper.driver.find_element(By.XPATH, '//a[@class="page-logo"]')
        page_logo.send_keys(Keys.CONTROL + Keys.ENTER) # ctrl+click -- opens the link in new background tab
        manual.confirm('you should not see any new notifications')
        # TODO switch to new tab?
        # TODO https://www.e-flux.com/journal/53/


# TODO shit disappears on chrome and present on firefox
@browsers()
def test_local_page(tmp_path: Path, browser: Browser) -> None:
    tutorial = 'file:///usr/share/doc/python3/html/tutorial/index.html'
    urls = {
         tutorial                                                : 'TODO read this',
        'file:///usr/share/doc/python3/html/reference/index.html': None,
    }
    url = PYTHON_DOC_URL
    with _test_helper(tmp_path, index_urls(urls), url, browser=browser) as helper:
        confirm('grey icon')
        helper.driver.get(tutorial)
        confirm('green icon. MANUALLY: ACTIVATE SIDEBAR!. It should open sidebar with one visit')
        helper.driver.back()
        # TODO it's always guaranteed to work? https://stackoverflow.com/questions/27626783/python-selenium-browser-driver-back
        confirm('grey icon, should be no sidebar')
        # fIXME fuck. failing here to load anchorme as well as webext-options-sync? ugh.
        helper.driver.forward()
        confirm('green icon, sidebar visible')


@browsers()
def test_unreachable(tmp_path, browser) -> None:
    url = 'https://somenonexist1ngurl.com'
    urls = {
        url: 'some context',
    }
    with _test_helper(
            tmp_path, index_urls(urls), 'about:blank', browser=browser,
            notify_contexts=True,
            verbose_errors=False,
    ) as helper:
        try:
            helper.driver.get(url)
        except:
            # results in exception because it's unreachable
            pass
        # TODO maybe in this case it could instead open the sidebar in a separate tab?
        manual.confirm('green icon, no errors, desktop notification with contexts')


@browsers()
def test_stress(tmp_path, browser) -> None:
    url = 'https://www.reddit.com/'
    urls = [
        (f'{url}/subpath/{i}.html', f'context {i}' if i > 10000 else None) for i in range(50000)
    ]
    with _test_helper(tmp_path, index_urls(urls), url, browser=browser) as helper:
        if has_x():
            helper.activate()

        manual.confirm('''
Is performance reasonable?
The sidebar should show up, and update gradually.
You should be able to scroll the page, trigger tooltips, etc., without any lags.
'''.strip())

    
@browsers(FF, CH)
def test_fuzz(tmp_path, browser):
    # TODO ugh. this still results in 'tab permissions' pages, but perhaps because of closed tabs?
    # dunno if it's worth fixing..
    urls = {
        'https://www.iana.org/domains/reserved': 'IANA',
        'iana.org/domains/reserved': 'IANA2',
    }
    with _test_helper(tmp_path, index_urls(urls), 'https://example.com', browser=browser) as helper:
        driver = helper.driver
        tabs = 30
        for _ in range(tabs):
            driver.find_element(By.TAG_NAME, 'a').send_keys(Keys.CONTROL + Keys.RETURN)

        sleep(5)
        for _ in range(tabs - 2):
            driver.close()
            sleep(0.1)
            driver.switch_to.window(driver.window_handles[0])

        def cb():
            for _ in range(10):
                send_key('Ctrl+Shift+t')
                sleep(0.1)
        trigger_callback(driver, cb)
        # FIXME shows quite a bunch of notifications here...
        confirm("shouldn't result in 'unexpected error occured'; show only show single notification per page")


def trigger_sidebar_search(driver):
    driver.switch_to.default_content()
    driver.switch_to.frame('promnesia-sidebar')
    search_button = driver.find_element_by_xpath('//button[text()="Search"]')
    search_button.click()


@browsers(FF, CH)
def test_duplicate_background_pages(tmp_path, browser):
    url = PYTHON_DOC_URL
    with _test_helper(tmp_path, index_urls({'whatever.coom': '123'}), url, browser=browser) as helper:
        driver = helper.driver

        trigger_command(driver, Command.ACTIVATE)
        # TODO separate test just for buttons from extension
        confirm('sidebar opened?')

        trigger_sidebar_search(driver)
        sleep(1)
        driver.switch_to.window(driver.window_handles[0])
        sleep(1)

        trigger_sidebar_search(driver)
        sleep(1)
        driver.switch_to.window(driver.window_handles[0])
        sleep(1)

        confirm('only two search pages should be opened')

        trigger_command(driver, Command.ACTIVATE)

        confirm('sidebar should be closed now')

        # TODO wtf? browser with search pages stays open after test... 

        # TODO getting this in chrome inspector while running this...
# VM2048 common.js:116 [background] [object Object]
# log @ VM2048 common.js:116
# notifyError @ VM2056 notifications.js:40
# Promise.catch (async)
# (anonymous) @ VM2056 notifications.js:49
# VM2056 notifications.js:17 Uncaught (in promise) TypeError: Cannot read property 'create' of undefined
#     at notify (VM2056 notifications.js:17)
#     at notifyError (VM2056 notifications.js:41)



# TODO FIXME need to test racey conditions _while_ page is loading, results in this 'unexpected error occured'?


# TODO shit, sometimes I have 'bindSidebarData is not defined'? with vebose errors on demo_how_did_i_get_here



import argparse
import logging
import inspect
import sys
from typing import List, Tuple, Optional, Dict, Sequence, Iterable, Iterator, Union
from pathlib import Path
from datetime import datetime
from .compat import check_call, register_argparse_extend_action_in_pre_py38
from tempfile import TemporaryDirectory


from . import config
from . import server
from .misc import install_server
from .common import PathIsh, logger, get_tmpdir, DbVisit, Res
from .common import Source, get_system_tz, user_config_file, default_config_path
from .dump import visits_to_sqlite
from .extract import extract_visits, make_filter


def iter_all_visits(sources_subset: Iterable[Union[str, int]]=()) -> Iterator[Res[DbVisit]]:
    cfg = config.get()
    output_dir = cfg.output_dir
    # not sure if belongs here??
    if not output_dir.exists():
        logger.warning("OUTPUT_DIR '%s' didn't exist, creating", output_dir)
        output_dir.mkdir(exist_ok=True, parents=True)

    hook = cfg.hook

    indexers = list(cfg.sources)

    is_subset_sources = bool(sources_subset)
    if is_subset_sources:
        sources_subset = set(sources_subset)

    for i, idx in enumerate(indexers):
        name = getattr(idx, "name", None)
        if name and is_subset_sources:
            matched = name in sources_subset or i in sources_subset
            if matched:
                sources_subset -= {i, name}  # type: ignore
            else:
                logger.debug("skipping '%s' not in --sources.", name)
                continue

        if isinstance(idx, Exception):
            yield idx
            continue
        # todo use this context? not sure where to attach...
        einfo = f'{getattr(idx.ff, "__module__", None)}:{getattr(idx.ff, "__name__", None)} {idx.args} {idx.kwargs}'
        for v in extract_visits(idx, src=idx.name):
            if hook is None:
                yield v
            else:
                try:
                    yield from hook(v)
                except Exception as e:
                    yield e

    if sources_subset:
        logger.warning("unknown --sources: %s", ", ".join(repr(i) for i in sources_subset))


def _do_index(dry: bool=False, sources_subset: Iterable[Union[str, int]]=(), overwrite_db=False) -> Iterable[Exception]:
    # also keep & return errors for further display
    errors: List[Exception] = []
    def it() -> Iterable[Res[DbVisit]]:
        for v in iter_all_visits(sources_subset):
            if isinstance(v, Exception):
                errors.append(v)
            yield v

    if dry:
        res = list(it())
        logger.warning("DRY MODE: won't modify the database. Printing the results out")
        for v in res:
            print(v)
    else:
        dump_errors = visits_to_sqlite(it(), overwrite_db=overwrite_db)
        for e in dump_errors:
            logger.exception(e)
            errors.append(e)
    return errors


def do_index(
        config_file: Path,
        dry: bool=False,
        sources_subset: Iterable[Union[str, int]]=(),
        overwrite_db: bool=False,
    ) -> None:
    config.load_from(config_file) # meh.. should be cleaner
    try:
        errors = list(_do_index(dry=dry, sources_subset=sources_subset, overwrite_db=overwrite_db))
    finally:
        config.reset()
    if len(errors) > 0:
        logger.error('%d errors, printing them out:', len(errors))
        for e in errors:
            logger.error('    %s', e)
        logger.error('%d errors, exit code 1', len(errors))
        sys.exit(1)


def demo_sources():
    def lazy(name: str):
        # helper to avoid failed imports etc, since people might be lacking necessary dependencies
        def inner(*args, **kwargs):
            from . import sources
            import importlib
            module = importlib.import_module('promnesia.sources' + '.' + name)
            return getattr(module, 'index')
        return inner

    res = {}
    import ast
    import promnesia.sources
    for p in promnesia.sources.__path__: # type: ignore[attr-defined] # should be present
        for x in sorted(Path(p).glob('*.py')):
            a = ast.parse(x.read_text())
            candidates = [c for c in a.body if getattr(c, 'name', None) == 'index']
            if len(candidates) > 0:
                res[x.stem] = lazy(x.stem)
    return res


def do_demo(
        *,
        index_as: str,
        params: Sequence[str],
        port: Optional[str],
        config_file: Optional[Path],
        dry: bool=False,
        name: str='demo',
        sources_subset: Iterable[Union[str, int]]=(),
        overwrite_db: bool=False,
    ) -> None:
    from pprint import pprint
    with TemporaryDirectory() as tdir:
        outdir = Path(tdir)

        if config_file is not None:
            config.load_from(config_file)
        else:
            idx = demo_sources()[index_as]()
            src = Source(idx, *params, name=name)
            cfg = config.Config(
                OUTPUT_DIR=outdir,
                SOURCES=[src],
            )
            config.instance = cfg

        errors = list(_do_index(dry=dry, sources_subset=sources_subset, overwrite_db=overwrite_db))
        if len(errors) > 0:
            logger.error('%d errors during indexing (see logs above for backtraces)', len(errors))
        for e in errors:
            logger.error(e)

        dbp = config.get().db
        if port is None:
            logger.warning(f"Port isn't specified, not serving!\nYou can inspect the database in the meantime, e.g. 'sqlitebrowser {dbp}'")
        else:
            from .server import ServerConfig
            server._run(
                port=port,
                quiet=False,
                config=ServerConfig(
                    db=dbp,
                    timezone=get_system_tz()
                ),
            )

        if sys.stdin.isatty():
            input("Press any key when ready")


def read_example_config() -> str:
    import inspect
    from .misc import config_example
    return inspect.getsource(config_example)


def config_create(args: argparse.Namespace) -> None:
    cfg = user_config_file()
    cfgdir = cfg.parent
    if cfgdir.exists():
        logger.error('Config directory %s already exists. Aborting', cfgdir)
        sys.exit(1)
    else:
        stub = read_example_config()
        cfgdir.mkdir(parents=True)
        cfg.write_text(stub)
        logger.info("Created a stub config in '%s'. Edit it to tune to your liking. (see https://github.com/karlicoss/promnesia#setup for more info)", cfg)


def config_check(args: argparse.Namespace) -> None:
    cfg: Path = args.config
    errors = list(_config_check(cfg))
    if len(errors) == 0:
        logger.info('OK')
    else:
        logger.error('CHECK FAILED')
        sys.exit(1)


def _config_check(cfg: Path) -> Iterable[Exception]:
    from .compat import run

    logger.info('config: %s', cfg)

    def check(cmd) -> Iterable[Exception]:
        logger.debug(' '.join(map(str, cmd)))
        res = run(cmd)
        if res.returncode > 0:
            yield Exception()

    logger.info('Checking syntax...')
    cmd = [sys.executable, '-m', 'compileall', cfg]
    yield from check(cmd)

    # todo not sure if should be more defensive than check_call here
    logger.info('Checking type safety...')
    try:
        import mypy
    except ImportError:
        logger.warning("mypy not found, can't use it to check config!")
    else:
        yield from check([
            sys.executable, '-m', 'mypy',
            '--namespace-packages',
            '--color-output', # not sure if works??
            '--pretty',
            '--show-error-codes',
            '--show-error-context',
            '--check-untyped-defs',
            cfg,
        ])

    logger.info('Checking runtime errors...')
    yield from check([sys.executable, cfg])


def cli_doctor_db(args: argparse.Namespace) -> None:
    # todo could fallback to 'sqlite3 <db> .dump'?
    config.load_from(args.config) # TODO meh
    db = config.get().db
    if not db.exists():
        logger.error("Database {db} doesn't exist!")
        sys.exit(1)
    else:
        logger.info(f'OK, database exists: {db}')

    cmd = ['sqlite3', str(db), 'select src, COUNT(*) from visits GROUP BY src']
    logger.info(f'Querying database summary: {cmd}')
    check_call(cmd)

    bro = 'sqlitebrowser'
    import shutil
    if not shutil.which(bro):
        logger.error(f'Install {bro} to inspect the database!')
        sys.exit(1)
    cmd = [bro, str(db)]
    logger.debug(f'Running {cmd}')
    from .compat import Popen
    Popen(cmd)


def cli_doctor_server(args: argparse.Namespace) -> None:
    port = args.port
    endpoint = f'http://localhost:{port}/status'
    cmd = ['curl', endpoint]
    logger.info(f'Running {cmd}')
    check_call(cmd)
    print() # curl doesn't add newline
    logger.info('You should see the database path and version above!')


def _ordinal_or_name(s: str) -> Union[str, int]:
    try:
        s = int(s)  # type: ignore
    except ValueError:
        pass
    return s


def main() -> None:
    # TODO longer, literate description?

    def add_index_args(parser: argparse.ArgumentParser, default_config_path: PathIsh=None) -> None:
        """
        :param default_config_path:
            if not given, all :func:`demo_sources()` are run
        """
        register_argparse_extend_action_in_pre_py38(parser)
        parser.add_argument('--config', type=Path, default=default_config_path, help='Config path')
        parser.add_argument('--dry', action='store_true', help="Dry run, won't touch the database, only print the results out")
        parser.add_argument(
            '--sources',
            required=False,
            action="extend",
            nargs="+",
            type=_ordinal_or_name,
            metavar="SOURCE",
            help="Source names (or their 0-indexed position) to index.",
        )
        parser.add_argument(
            '--overwrite',
            required=False,
            action="store_true",
            help="Empty db before populating it with newly indexed visits."
            "  If interrupted, db is left untouched."
        )

    F = lambda prog: argparse.ArgumentDefaultsHelpFormatter(prog, width=120)
    p = argparse.ArgumentParser(formatter_class=F) # type: ignore
    subp = p.add_subparsers(dest='mode', )
    ep = subp.add_parser('index', help='Create/update the link database', formatter_class=F)
    add_index_args(ep, default_config_path())
    # TODO use some way to override or provide config only via cmdline?
    ep.add_argument('--intermediate', required=False, help="Used for development, you don't need it")

    sp = subp.add_parser('serve', help='Serve a link database', formatter_class=F) # type: ignore
    server.setup_parser(sp)

    ap = subp.add_parser('demo', help='Demo mode: index and serve a directory in single command', formatter_class=F)
    # TODO use docstring or something?
    #

    add_port_arg = lambda p: p.add_argument('--port', type=str, default='13131'              , help='Port to serve on')

    ap.add_argument('--name', type=str, default='demo'               , help='Set custom source name')
    add_port_arg(ap)
    ap.add_argument('--no-serve', action='store_const', const=None, dest='port', help='Pass to only index without running server')
    ap.add_argument(
        '--as',
        choices=list(sorted(demo_sources().keys())),
        default='guess',
        help='Promnesia source to index as (see https://github.com/karlicoss/promnesia/tree/master/src/promnesia/sources for the full list)',
    )
    add_index_args(ap)
    ap.add_argument('params', nargs='*', help='Optional extra params for the indexer')

    isp = subp.add_parser('install-server', help='Install server as a systemd service (for autostart)', formatter_class=F)
    install_server.setup_parser(isp)

    cp = subp.add_parser('config', help='Config management')
    cp.set_defaults(func=lambda *args: cp.print_help())
    scp = cp.add_subparsers()
    ccp = scp.add_parser('check', help='Check config')
    ccp.set_defaults(func=config_check)
    ccp.add_argument('--config', type=Path, default=default_config_path(), help='Config path')

    icp = scp.add_parser('create', help='Create user config')
    icp.add_argument(
        "--config", type=Path, default=default_config_path(), help="Config path"
    )
    icp.set_defaults(func=config_create)

    dp = subp.add_parser('doctor', help='Troubleshooting assistant')
    dp.add_argument('--config', type=Path, default=default_config_path(), help='Config path')
    dp.set_defaults(func=lambda *args: dp.print_help())
    sdp = dp.add_subparsers()
    sdp.add_parser('config'  , help='Check config'    ).set_defaults(func=config_check )
    sdp.add_parser('database', help='Inspect database').set_defaults(func=cli_doctor_db)
    sdps = sdp.add_parser('server'  , help='Check server'    )
    sdps.set_defaults(func=cli_doctor_server)
    add_port_arg(sdps)

    args = p.parse_args()

    # TODO is there a way to print full help? i.e. for all subparsers
    if args.mode is None:
        print('ERROR: Please specify a mode', file=sys.stderr)
        p.print_help(sys.stderr)
        sys.exit(1)

    logger.info("CLI args: %s", args)

    # TODO maybe, it's better for server to compute intermediate representations?
    # the only downside is storage. dunno.
    # worst case -- could use database?

    with get_tmpdir() as tdir: # TODO??
        if args.mode == 'index':
            do_index(
                config_file=args.config,
                dry=args.dry,
                sources_subset=args.sources,
                overwrite_db=args.overwrite,
            )
        elif args.mode == 'serve':
            server.run(args)
        elif args.mode == 'demo':
            # TODO not sure if 'as' is that useful
            # something like Telegram/Takeout is too hard to setup to justify adhoc mode like this?
            do_demo(
                index_as=getattr(args, 'as'),
                params=args.params,
                port=args.port,
                config_file=args.config,
                dry=args.dry,
                name=args.name,
                sources_subset=args.sources,
                overwrite_db=args.overwrite,
                )
        elif args.mode == 'install-server': # todo rename to 'autostart' or something?
            install_server.install(args)
        elif args.mode == 'config':
            args.func(args)
        elif args.mode == 'doctor':
            args.func(args)
        else:
            raise AssertionError(f'unexpected mode {args.mode}')

if __name__ == '__main__':
    main()
